
import datetime
import pytest
from rocketry.time.interval import (
    TimeOfMinute,
    TimeOfDay,
    TimeOfHour,
    TimeOfMonth,
    TimeOfSecond,
    TimeOfWeek,
    TimeOfYear
)

MS_IN_MILLISECOND = 1000
MS_IN_SECOND = int(1e+6)
MS_IN_MINUTE = int(1e+6 * 60)
MS_IN_HOUR   = int(1e+6 * 60 * 60)
MS_IN_DAY    = int(1e+6 * 60 * 60 * 24)

def pytest_generate_tests(metafunc):
    if metafunc.cls is not None:

        method_name = metafunc.function.__name__
        cls = metafunc.cls
        params = []
        if method_name == "test_closed":
            params = cls.scen_closed
        elif method_name == "test_open":
            return
        elif method_name == "test_open_left":
            params = cls.scen_open_left
        elif method_name == "test_open_right":
            params = cls.scen_open_right
        elif method_name == "test_time_point":
            params = cls.scen_time_point
        elif method_name == "test_starting":
            params = cls.scen_starting
        elif method_name == "test_value_error":
            params = cls.scen_value_error
        else:
            return

        argnames = [arg for arg in metafunc.fixturenames if arg in ('start', 'end', "expected_start", "expected_end", "time_point")]

        idlist = []
        argvalues = []
        argvalues = []
        for scen in params:
            idlist.append(scen.pop("id", f'{repr(scen.get("start"))}, {repr(scen.get("end"))}'))
            argvalues.append(tuple(scen[name] for name in argnames))
        metafunc.parametrize(argnames, argvalues, ids=idlist, scope="class")

class ConstructTester:

    def test_closed(self, start, end, expected_start, expected_end):
        time = self.cls(start, end)
        assert not time.is_full()
        assert expected_start == time._start
        assert expected_end == time._end

    def test_open(self):
        time = self.cls(None, None)
        assert time.is_full()
        assert 0 == time._start
        assert self.max_ms == time._end

    def test_open_left(self, end, expected_end, **kwargs):
        time = self.cls(None, end)
        assert not time.is_full()
        assert 0 == time._start
        assert expected_end == time._end

    def test_open_right(self, start, expected_start, **kwargs):
        time = self.cls(start, None)
        assert not time.is_full()
        assert expected_start == time._start
        assert self.max_ms == time._end

    def test_time_point(self, start, expected_start, expected_end, **kwargs):
        time = self.cls(start, time_point=True)
        assert not time.is_full()
        assert expected_start == time._start
        assert expected_end == time._end

        time = self.cls.at(start)
        assert not time.is_full()
        assert expected_start == time._start
        assert expected_end == time._end

    def test_starting(self, start, expected_start, **kwargs):
        time = self.cls(start, starting=True)
        assert time.is_full()
        assert expected_start == time._start
        assert expected_start == time._end

        time = self.cls.starting(start)
        assert time.is_full()
        assert expected_start == time._start
        assert expected_start == time._end

    def test_value_error(self, start, end):
        with pytest.raises(ValueError):
            time = self.cls(start, end)

class TestTimeOfSecond(ConstructTester):

    cls = TimeOfSecond

    max_ms = MS_IN_SECOND

    scen_closed = [
        {
            "start": "15.005",
            "end": "45.005",
            "expected_start": 15 * MS_IN_MILLISECOND + 5,
            "expected_end": 45 * MS_IN_MILLISECOND + 5,
        },
       {
            "start": 15,
            "end": 45,
            "expected_start": 15 * MS_IN_MILLISECOND,
            "expected_end": 46 * MS_IN_MILLISECOND,
        },
       {
            "start": 15.005,
            "end": 45.005,
            "expected_start": 15 * MS_IN_MILLISECOND + 5,
            "expected_end": 45 * MS_IN_MILLISECOND + 5,
        },
    ]

    scen_open_left = [
        {
            "end": 45,
            "expected_end": 46 * MS_IN_MILLISECOND
        }
    ]
    scen_open_right = [
        {
            "start": 45,
            "expected_start": 45 * MS_IN_MILLISECOND
        }
    ]
    scen_time_point = [
        {
            "start": 500,
            "expected_start": 500 * MS_IN_MILLISECOND,
            "expected_end": 501 * MS_IN_MILLISECOND,
        }
    ]
    scen_starting = [
        {
            "start": 500,
            "expected_start": 500 * MS_IN_MILLISECOND,
        }
    ]
    scen_value_error = [
        {
            "start": 1001,
            "end": None
        },
        {
            "start": 1000.001,
            "end": None
        },
        {
            "start": "1001",
            "end": None
        },
        {
            "start": "asd",
            "end": None
        },
    ]

class TestTimeOfMinute(ConstructTester):

    cls = TimeOfMinute

    max_ms = MS_IN_MINUTE

    scen_closed = [
        {
            "start": "15",
            "end": "45",
            "expected_start": 15 * MS_IN_SECOND,
            "expected_end": 45 * MS_IN_SECOND,
        },
        {
            "start": "15.000",
            "end": "45.000",
            "expected_start": 15 * MS_IN_SECOND,
            "expected_end": 45 * MS_IN_SECOND,
        },
        {
            "start": "15.5",
            "end": "45.5",
            "expected_start": 15 * MS_IN_SECOND + MS_IN_MILLISECOND * 500,
            "expected_end": 45 * MS_IN_SECOND + MS_IN_MILLISECOND * 500,
        },
        {
            "start": "15.005",
            "end": "45.005",
            "expected_start": 15 * MS_IN_SECOND + MS_IN_MILLISECOND * 5,
            "expected_end": 45 * MS_IN_SECOND + MS_IN_MILLISECOND * 5,
        },
        {
            "start": "15.000005",
            "end": "45.000005",
            "expected_start": 15 * MS_IN_SECOND + 5,
            "expected_end": 45 * MS_IN_SECOND + 5,
        },
        {
            "start": 15,
            "end": 45,
            "expected_start": 15 * MS_IN_SECOND,
            "expected_end": 46 * MS_IN_SECOND,
        },
        {
            "start": 15.005,
            "end": 45.005,
            "expected_start": 15 * MS_IN_SECOND + 5 * MS_IN_MILLISECOND,
            "expected_end": 45 * MS_IN_SECOND + 5 * MS_IN_MILLISECOND,
        },
    ]

    scen_open_left = [
        {
            "end": "45.00",
            "expected_end": 45 * MS_IN_SECOND
        }
    ]
    scen_open_right = [
        {
            "start": "45",
            "expected_start": 45 * MS_IN_SECOND
        }
    ]
    scen_time_point = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_SECOND,
            "expected_end": 13 * MS_IN_SECOND,
        }
    ]
    scen_starting = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_SECOND,
        }
    ]
    scen_value_error = [
        {
            "start": 60,
            "end": None
        }
    ]

class TestTimeOfHour(ConstructTester):

    cls = TimeOfHour

    max_ms = MS_IN_HOUR

    scen_closed = [
        {
            "start": "15:00",
            "end": "45:00",
            "expected_start": 15 * MS_IN_MINUTE,
            "expected_end": 45 * MS_IN_MINUTE,
        },
        {
            "start": 15,
            "end": 45,
            "expected_start": 15 * MS_IN_MINUTE,
            "expected_end": 46 * MS_IN_MINUTE,
        },
        {
            "start": "15:05.5",
            "end": "45:10.005",
            "expected_start": 15 * MS_IN_MINUTE + 5 * MS_IN_SECOND + MS_IN_MILLISECOND * 500,
            "expected_end": 45 * MS_IN_MINUTE + 10 * MS_IN_SECOND + MS_IN_MILLISECOND * 5,
        },
    ]

    scen_open_left = [
        {
            "end": "45:00",
            "expected_end": 45 * MS_IN_MINUTE
        }
    ]
    scen_open_right = [
        {
            "start": "45:00",
            "expected_start": 45 * MS_IN_MINUTE
        }
    ]
    scen_time_point = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_MINUTE,
            "expected_end": 13 * MS_IN_MINUTE,
        }
    ]
    scen_starting = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_MINUTE,
        }
    ]

    scen_value_error = [
        {
            "start": 60,
            "end": None
        },
        {
            "start": None,
            "end": "60:01"
        },
        {
            # Float makes not much sense
            "start": 2.5,
            "end": None
        },
        {
            "start": "invalid",
            "end": None
        },
    ]

class TestTimeOfDay(ConstructTester):

    cls = TimeOfDay

    max_ms = 24 * MS_IN_HOUR

    scen_closed = [
        {
            "start": "10:00",
            "end": "12:00",
            "expected_start": 10 * MS_IN_HOUR,
            "expected_end": 12 * MS_IN_HOUR,
        },
        {
            "start": 10,
            "end": 12,
            "expected_start": 10 * MS_IN_HOUR,
            "expected_end": 13 * MS_IN_HOUR,
        },
        {
            "start": {"hour": 10, "second": 20},
            "end": {"hour": 13, "second": 20},
            "expected_start": 10 * MS_IN_HOUR + 20 * MS_IN_SECOND,
            "expected_end": 13 * MS_IN_HOUR + 20 * MS_IN_SECOND,
        },
    ]

    scen_open_left = [
        {
            "end": "12:00",
            "expected_end": 12 * MS_IN_HOUR
        }
    ]
    scen_open_right = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_HOUR
        }
    ]
    scen_time_point = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_HOUR,
            "expected_end": 13 * MS_IN_HOUR,
        },
        {
            "start": "23:00",
            "expected_start": 23 * MS_IN_HOUR,
            "expected_end": 0,
        },
        {
            "start": "23:01",
            "expected_start": 23 * MS_IN_HOUR + MS_IN_MINUTE,
            "expected_end": MS_IN_MINUTE,
        },
    ]
    scen_starting = [
        {
            "start": "12:00",
            "expected_start": 12 * MS_IN_HOUR,
        }
    ]
    scen_value_error = [
        {
            "start": -1,
            "end": None,
        },
        {
            "start": "asd",
            "end": None,
        },
        {
            "start": 24,
            "end": None,
        },
        {
            "start": None,
            "end": "24:01",
        },
        {
            # Float makes not much sense
            "start": 2.5,
            "end": None
        },
    ]


class TestTimeOfWeek(ConstructTester):

    cls = TimeOfWeek

    max_ms = 7 * MS_IN_DAY

    scen_closed = [
        {
            # Spans from Tue 00:00:00 to Wed 23:59:59 999
            "start": "Tue",
            "end": "Wed",
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 3 * MS_IN_DAY,
        },
        {
            # Spans from Tue 00:00:00 to Wed 23:59:59 999
            "start": "Tuesday",
            "end": "Wednesday",
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 3 * MS_IN_DAY,
        },
        {
            # Spans from Tue 00:00:00 to Wed 23:59:59 999
            "start": 2,
            "end": 3,
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 3 * MS_IN_DAY,
        },
    ]

    scen_open_left = [
        {
            "end": "Tue",
            "expected_end": 2 * MS_IN_DAY # Tuesday 23:59:59 ...
        }
    ]
    scen_open_right = [
        {
            "start": "Tue",
            "expected_start": 1 * MS_IN_DAY # Tuesday 00:00:00
        }
    ]
    scen_time_point = [
        {
            "start": "Tue",
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 2 * MS_IN_DAY,
        }
    ]
    scen_starting = [
        {
            "start": "Tue",
            "expected_start": 1 * MS_IN_DAY,
        }
    ]
    scen_value_error = [
        {
            "start": 0,
            "end": None
        },
        {
            "start": "Asd",
            "end": None
        },
        {
            # Float makes not much sense
            "start": 2.5,
            "end": None
        },
    ]


class TestTimeOfMonth(ConstructTester):

    cls = TimeOfMonth

    max_ms = 31 * MS_IN_DAY

    scen_closed = [
        {
            "start": "2.",
            "end": "3.",
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 3 * MS_IN_DAY,
        },
        {
            "start": "2nd",
            "end": "4th",
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 4 * MS_IN_DAY,
        },
        {
            "start": 2,
            "end": 4,
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 4 * MS_IN_DAY,
        },
    ]

    scen_open_left = [
        {
            "end": "3.",
            "expected_end": 3 * MS_IN_DAY
        }
    ]
    scen_open_right = [
        {
            "start": "2.",
            "expected_start": 1 * MS_IN_DAY
        }
    ]
    scen_time_point = [
        {
            "start": "2.",
            "expected_start": 1 * MS_IN_DAY,
            "expected_end": 2 * MS_IN_DAY,
        }
    ]
    scen_starting = [
        {
            "start": "2.",
            "expected_start": 1 * MS_IN_DAY,
        }
    ]
    scen_value_error = [
        {
            "start": 0,
            "end": None,
        },
        {
            "start": None,
            "end": 32,
        },
        {
            "start": "33.",
            "end": None
        },
        {
            # Float makes not much sense
            "start": 2.5,
            "end": None
        },
    ]

class TestTimeOfYear(ConstructTester):

    cls = TimeOfYear

    max_ms = 366 * MS_IN_DAY # Leap year has 366 days

    scen_closed = [
        {
            "start": "February",
            "end": "April",
            "expected_start": 31 * MS_IN_DAY,
            "expected_end": (31 + 29 + 31 + 30) * MS_IN_DAY - 1,
        },
        {
            "start": "Feb",
            "end": "Apr",
            "expected_start": 31 * MS_IN_DAY,
            "expected_end": (31 + 29 + 31 + 30) * MS_IN_DAY - 1,
        },
        {
            "start": 2,
            "end": 4,
            "expected_start": 31 * MS_IN_DAY,
            "expected_end": (31 + 29 + 31 + 30) * MS_IN_DAY - 1,
        },
    ]

    scen_open_left = [
        {
            "end": "Apr",
            "expected_end": (31 + 29 + 31 + 30) * MS_IN_DAY - 1
        },
        {
            "end": "Jan",
            "expected_end": 31 * MS_IN_DAY - 1
        },
    ]
    scen_open_right = [
        {
            "start": "Apr",
            "expected_start": (31 + 29 + 31) * MS_IN_DAY
        },
        {
            "start": "Dec",
            "expected_start": (366 - 31) * MS_IN_DAY
        },
    ]
    scen_time_point = [
        {
            "start": "Jan",
            "expected_start": 0,
            "expected_end": 31 * MS_IN_DAY - 1,
        },
        {
            "start": "Feb",
            "expected_start": 31 * MS_IN_DAY,
            "expected_end": (31 + 29) * MS_IN_DAY - 1,
        },
        {
            "start": "Dec",
            "expected_start": (366 - 31) * MS_IN_DAY,
            "expected_end": 366 * MS_IN_DAY - 1,
        },
    ]
    scen_starting = [
        {
            "start": "Feb",
            "expected_start": 31 * MS_IN_DAY,
        }
    ]
    scen_value_error = [
        {
            "start": 0,
            "end": None,
        },
        {
            "start": None,
            "end": 13,
        },
        {
            # Float makes not much sense
            "start": 2.5,
            "end": None
        },
    ]

@pytest.mark.parametrize("cls", [TimeOfSecond, TimeOfMinute, TimeOfHour, TimeOfDay])
def test_type_error(cls):
    with pytest.raises(TypeError):
        time = cls.starting(lambda: None)
    with pytest.raises(TypeError):
        time = cls.starting(datetime.datetime(2022, 1, 1))
    with pytest.raises(TypeError):
        time = cls.starting(datetime.timedelta(days=2))

    with pytest.raises(ValueError):
        time = cls(time_point=True)



from rocketry.core import BaseCondition, Task

from rocketry.session import Session, Config
from rocketry.parse import add_condition_parser
from rocketry.conds import true, false
from rocketry.tasks import CommandTask, FuncTask, CodeTask
from rocketry.tasks.maintain import ShutDown, Restart

from rocketry.conditions.meta import _FuncTaskCondWrapper

def _setup_defaults():
    "Set up the task classes and conditions Rocketry provides out-of-the-box"

    # Add some extra parsers from core
    add_condition_parser({
        "true": true,
        "false": false,
        "always false": false,
        "always true": true
    })

    # Update type hints
    cls_tasks = (
        Task,
        FuncTask, CommandTask, CodeTask,
        ShutDown, Restart,

        _FuncTaskCondWrapper
    )
    for cls_task in cls_tasks:
        cls_task.update_forward_refs(Session=Session, BaseCondition=BaseCondition)

    Config.update_forward_refs(BaseCondition=BaseCondition)
    #Session.update_forward_refs(
    #    Task=Task, Parameters=Parameters, Scheduler=Scheduler
    #)



from typing import Any, TYPE_CHECKING
from abc import abstractmethod

from rocketry._base import RedBase

if TYPE_CHECKING:
    import rocketry

class BaseArgument(RedBase):
    """Base class for Arguments.

    Argument is a wrapper for value that can be
    passed to a task. The value can be formulated
    on the fly or it can be a constant variable.

    Examples
    --------

    Minimum example:

    >>> from rocketry.core import BaseArgument
    >>> class MyArgument(BaseArgument):
    ...     def __init__(self, value):
    ...         self.value = value
    ...
    ...     def get_value(self):
    ...         value = self.value
    ...         ... # Code that forms the argument value
    ...         return value
    ...
    >>> from rocketry.core import Parameters
    >>> Parameters(myarg=MyArgument("value"))
    Parameters(myarg=MyArgument('value'))

    """
    session: 'rocketry.Session'

    @abstractmethod
    def get_value(self, **kwargs) -> Any:
        """Get the actual value of the argument.
        Override for custom behaviour.

        Parameters
        ----------
        task : rocketry.core.Task, optional
            Task that requested the value of the argument.
        """

    def stage(self, **kwargs) -> 'BaseArgument':
        """Get (a copy of) the argument with a
        value that can be passed to child threads
        or processes. Override for custom behaviour.

        Example use cases:

            - Save a complex object to disk and read
              it again later with ``get_value``.
            - Shut a file buffer and reopen it when
              in the child process/thread.
            - Remove non-picklable attributes.

        Parameters
        ----------
        task : rocketry.core.Task, optional
            Task that requested the value of the argument.

        Notes
        -----
        It is advisable to copy the argument in order
        not to interfere with the passing of the argument
        in later situations.

        Returns
        -------
        BaseArgument
            Argument that is ready to be passed to a child
            thread or process.
        """
        return self.get_value(**kwargs)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.get_value() == other.get_value()
        return False

    def __repr__(self):
        cls_name = type(self).__name__
        return f'{cls_name}({repr(self.get_value())})'

    def __str__(self):
        return str(self.get_value())

    def __rshift__(self, other):
        args = []
        if not isinstance(other, BaseArgument):
            raise TypeError(f"Invalid type {type(other)}")
        for inst in (self, other):
            if isinstance(inst, PipeArg):
                args += list(inst)
            else:
                args.append(inst)
        return PipeArg(args)

class PipeArg(BaseArgument):

    def __init__(self, args:list):
        self._args = args

    def __iter__(self):
        return iter(self._args)

    def get_value(self, **kwargs) -> Any:
        for arg in self:
            try:
                return arg.get_value(**kwargs)
            except KeyError:
                continue
        else:
            raise KeyError("No value found")



"""
Utilities for getting information
about the scehuler/task/parameters etc.
"""

import datetime
import logging
from multiprocessing import cpu_count
import warnings

from itertools import chain
from typing import TYPE_CHECKING, Callable, ClassVar, Iterable, Dict, List, Optional, Set, Tuple, Union
from pydantic import BaseModel, validator
from rocketry.pybox.time import to_timedelta
from rocketry.log.defaults import create_default_handler
from rocketry._base import RedBase


try:
    from typing import Literal
except ImportError: # pragma: no cover
    from typing_extensions import Literal


if TYPE_CHECKING:
    from rocketry.core.log import TaskAdapter
    from rocketry.parse import StaticParser
    from rocketry.core import (
        Task,
        Scheduler,
        BaseCondition,
        Parameters,
        BaseArgument,
        TimePeriod
    )



class Config(BaseModel):
    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    # Fields
    use_instance_naming: bool = False
    task_priority: int = 0
    task_execution: Optional[str] = None
    task_pre_exist: str = 'raise'
    force_status_from_logs: bool = False # Force to check status from logs every time (slow but robust)

    task_logger_basename: str = "rocketry.task"
    scheduler_logger_basename: str = "rocketry.scheduler"

    silence_task_prerun: bool = False # Whether to silence errors occurred in setting a task to run
    silence_task_logging: bool = False # Whether to silence errors occurred in logging a task
    silence_cond_check: bool = False # Whether to silence errors occurred in checking conditions
    cycle_sleep: Optional[float] = 0.1
    debug: bool = False

    max_process_count = cpu_count()
    tasks_as_daemon: bool = True
    restarting: str = 'replace'
    instant_shutdown: bool = False

    timeout: datetime.timedelta = datetime.timedelta(minutes=30)
    shut_cond: Optional['BaseCondition'] = None

    param_materialize:Literal['pre', 'post'] = 'post'

    @validator('task_execution', pre=True, always=True)
    def parse_task_execution(cls, value):
        if value is None:
            warnings.warn(
                "Default execution will be changed to 'async'. "
                "To suppress this warning, specify task_execution, "
                "ie. Rocketry(execution='async')",
                FutureWarning
            )
            return 'process'
        return value

    @validator('shut_cond', pre=True)
    def parse_shut_cond(cls, value):
        from rocketry.parse import parse_condition
        from rocketry.conditions import AlwaysFalse
        if value is None:
            return AlwaysFalse()
        return parse_condition(value)

    @validator('timeout', pre=True, always=True)
    def parse_timeout(cls, value):
        if isinstance(value, str):
            return to_timedelta(value)
        elif isinstance(value, (float, int)):
            return datetime.timedelta(seconds=value)
        else:
            return value

class Hooks(BaseModel):
    task_init: List[Callable] = []
    task_execute: List[Callable] = []

    scheduler_startup: List[Callable] = []
    scheduler_cycle: List[Callable] = []
    scheduler_shutdown: List[Callable] = []

class Session(RedBase):
    """Collection of the scheduler objects.

    Parameters
    ----------

    config : dict, optional
        Central configuration for defining behaviour
        of different object and classes in the session.
    tasks : Dict[str, rocketry.core.Task], optional
        Tasks of the session. Can be formed later.
    parameters : parameter-like, optional
        Session level parameters.
    scheme : str or list, optional
        Premade scheme(s) to use to set up logging,
        parameters, setup tasks etc.
    as_default : bool, default=True
        Whether to set the session as default for next
        tasks etc. that don't have session
        specified.
    kwds_scheduler : dict, optional
        Keyword arguments passed to
        :py:class:`rocketry.core.Scheduler`.
    delete_existing_loggers : bool, default=False
        If True, deletes the loggers that already existed
        for the task logger basename.

    Attributes
    ----------
    config : dict
        Central configuration for defining behaviour
        of different object and classes in the session.
    scheduler : Scheduler
        Scheduler of the session.
    delete_existing_loggers : bool
        If True, all loggers that match the
        session.config.basename are deleted (by
        default, deletes loggers starting with
        'rocketry.task').

    """
    config: Config
    class Config:
        arbitrary_types_allowed = True

    tasks: Set['Task']
    hooks: Hooks
    parameters: 'Parameters'
    _scheduler: 'Scheduler'

    _time_parsers: ClassVar[Dict] = {}
    _cls_cond_parsers: ClassVar[Dict] = {} # Default condition parsers

    def _get_parameters(self, value):
        from rocketry.core import Parameters
        if value is None:
            return Parameters()
        elif not isinstance(value, Parameters):
            value = Parameters(value)
        return value

    def _get_config(self, value):
        if value is None:
            return Config()
        elif isinstance(value, dict):
            return Config(**value)
        elif isinstance(value, Config):
            return value
        else:
            raise TypeError("Invalid config type")

    @staticmethod
    def _get_task_name(task):
        from rocketry.core import Task
        if isinstance(task, str):
            task_name = task
        elif hasattr(task, "__rocketry__"):
            # Function that FuncTask set the rocketry info
            task_name = task.__rocketry__['name']
        elif isinstance(task, Task):
            task_name = task.name
        else:
            raise TypeError(f"Cannot determine task name from: {type(task)}")
        return task_name

    def __init__(self, config=None, parameters=None, delete_existing_loggers=False):
        from rocketry.core import Scheduler
        self.config = self._get_config(config)
        self.parameters = self._get_parameters(parameters)
        self.scheduler = Scheduler(self)
        self.tasks = set()
        self.hooks = Hooks()
        self.returns = self._get_parameters(None)
        self._cond_parsers = self._cls_cond_parsers.copy()
        self._cond_cache: Dict = {} # Cached by CondParser to speed up expensive conditions
        self._cond_states = {} # Used by FuncConds to relay condiiton states to conditions
        if delete_existing_loggers:
            self.delete_task_loggers()

    def __getitem__(self, task:Union['Task', str]):
        "Get a task from the session"
        task_name = self._get_task_name(task)
        for task in self.tasks:
            if task.name == task_name:
                return task
        else:
            raise KeyError(f"Task '{task_name}' not found")

    def __contains__(self, task: Union['Task', str]):
        "Check if task is in session"
        try:
            self[task]
        except KeyError:
            return False
        else:
            return True

    def start(self):
        """Start the scheduling session.

        Will block and wait till the scheduler finishes
        if there is a shut condition."""
        self._check_readable_logger()
        self.scheduler()

    async def serve(self):
        """Start the scheduling session using async.

        Will block and wait till the scheduler finishes
        if there is a shut condition."""
        self._check_readable_logger()
        await self.scheduler.serve()

    def run(self, *task_names:Tuple[str], execution=None, obey_cond=False):
        """Run specific task(s) manually.

        This method starts up the scheduler but only the given
        task is run. Useful to manually run a task while using
        the setup/teardown and parameters of the session and
        scheduler.

        Parameters
        ----------
        *task_names : tuple of str
            Names of the tasks to run.
        execution : str
            Execution method for all of the tasks.
            By default, whatever set to each task
        obey_cond : bool
            Whether to obey the ``start_cond`` or
            force a run regardless. By default, False

        .. warning::

            This is not meant to be called by tasks or the system
            itself. Just to run specific tasks when the system itself
            is not running.
        """
        self._check_readable_logger()
        # To prevent circular import
        from rocketry.conditions.scheduler import SchedulerCycles

        orig_vals = {}
        for task in self.tasks:
            name = task.name
            orig_vals[name] = {
                attr: val for attr, val in task.__dict__.items()
                if attr not in ("status", "last_run", "last_success", "last_fail", "last_terminate")
            }
            if name in task_names:
                if not obey_cond:
                    task.run()
                if execution is not None:
                    task.execution = execution
            else:
                task.disabled = True

        orig_shut_cond = self.config.shut_cond
        try:
            self.config.shut_cond = SchedulerCycles() >= 1
            self.start()
        finally:
            self.config.shut_cond = orig_shut_cond
            # Set back the disabled, execution etc.
            for task in self.tasks:
                task.__dict__.update(orig_vals[task.name])

    def restart(self):
        """Restart the scheduler

        The restart is not instantenous and
        will occur after the scheduler finishes
        checking one cycle of tasks."""
        self.scheduler._flag_restart.set()

    def shutdown(self):
        """Shut down the scheduler

        The shut down is not instantenous and
        will occur after the scheduler finishes
        checking one cycle of tasks."""
        warnings.warn((
            "Session.shutdown is deprecated. "
            "Please use Session.shut_down instead"
        ), DeprecationWarning)
        self.scheduler._flag_shutdown.set()

    def shut_down(self, force=None):
        """Shut down the scheduler"""
        force = force if force is not None else self.scheduler._flag_shutdown.is_set()
        self.scheduler._flag_shutdown.set()
        if force:
            self.scheduler._flag_force_exit.set()

    def _check_readable_logger(self):
        from rocketry.core.log import TaskAdapter
        task_logger_basename = self.config.task_logger_basename
        task_logger = logging.getLogger(task_logger_basename)
        logger = TaskAdapter(task_logger, None, ignore_warnings=True)
        if logger.is_readable_unset:
            # Setting memory logger
            warnings.warn(
                f"Logger {task_logger_basename} cannot be read. "
                "Logging is set to memory. "
                "To supress this warning, "
                "please set a handler that can be read (redbird.logging.RepoHandler)", UserWarning)

            # Setting memory logger
            task_logger.addHandler(create_default_handler())
        is_info_logged = logger.getEffectiveLevel() <= logging.INFO
        if not is_info_logged:
            level_name = logging.getLevelName(task_logger.getEffectiveLevel())
            warnings.warn(
                f"Logger {task_logger_basename} has too low level ({level_name}). "
                "Level is set to INFO to make sure the task logs get logged. ", UserWarning)
            task_logger.setLevel(logging.INFO)

    def get_tasks(self) -> list:
        """Get session tasks as list.

        Returns
        -------
        list[rocketry.core.Task]
            List of tasks in the session.
        """
        return self.tasks

    def get_task(self, task):
        warnings.warn((
            "Method get_task will be removed in the future version."
            "Please use instead: session['task name']"
        ), DeprecationWarning)
        return self[task]

    def get_cond_parsers(self):
        "Used by the actual string condition parser"
        return self._cond_parsers

    def create_task(self, *, command=None, path=None, **kwargs):
        "Create a task and put it to the session"

        # To avoid circular imports
        from rocketry.tasks import CommandTask, FuncTask

        kwargs['session'] = self

        if command is not None:
            return CommandTask(command=command, **kwargs)
        elif path is not None:
            # Non-wrapped FuncTask
            return FuncTask(path=path, **kwargs)
        else:
            return FuncTask(name_include_module=False, _name_template='{func_name}', **kwargs)

    def add_task(self, task: 'Task'):
        "Add the task to the session"
        if_exists = self.config.task_pre_exist
        exists = task in self
        if exists:
            if if_exists == 'ignore':
                return
            elif if_exists == 'replace':
                self.tasks.remove(task)
                self.tasks.add(task)
            elif if_exists == 'raise':
                raise KeyError(f"Task '{task.name}' already exists")
        else:
            self.tasks.add(task)

        # Adding the session to the task
        task.session = self

    def remove_task(self, task: Union['Task', str]):
        if isinstance(task, str):
            task = self[task]
        self.session.tasks.remove(task)

    def task_exists(self, task: 'Task'):
        warnings.warn((
            "Method task_exists will be removed in the future version."
            "Please use instead: 'task name' in session"
        ), DeprecationWarning)

        task_name = self._get_task_name(task)
        for task in self.tasks:
            if task.name == task_name:
                return True
        else:
            return False

    def get_repo(self):
        "Get log repo where the task logs are stored"
        from rocketry.core.log import TaskAdapter
        basename = self.config.task_logger_basename
        logger = logging.getLogger(basename)
        return TaskAdapter(logger, task=None)._get_repo()

    def get_task_loggers(self, with_adapters=True) -> Dict[str, Union['TaskAdapter', logging.Logger]]:
        """Get task logger(s) from the session.

        Parameters
        ----------
        with_adapters : bool, optional
            Whether get the loggers wrapped to
            rocketry.core.log.TaskAdapter, by default True

        Returns
        -------
        Dict[str, Union[TaskAdapter, logging.Logger]]
            Dictionary of the loggers (or adapters)
            in which the key is the logger name.
            Placeholders and loggers built for parallelized
            tasks are ignored.
        """
        from rocketry.core.log import TaskAdapter

        basename = self.config.task_logger_basename
        return {
            # The adapter should not be used to log (but to read) thus task_name = None
            name: TaskAdapter(logger, None) if with_adapters else logger
            for name, logger in logging.root.manager.loggerDict.items()
            if name.startswith(basename)
            and not isinstance(logger, logging.PlaceHolder)
            and not name.endswith("_process") # No private
        }

# Log data
    def get_task_log(self, *args, **kwargs) -> Iterable[Dict]:
        """Get task log records from all of the
        readable handlers in the session.

        Parameters
        ----------
        **kwargs : dict
            Query parameters passed to
            rocketry.core.log.TaskAdapter.get_records

        Returns
        -------
        Iterable[Dict]
            Generator of the task log records.
        """
        loggers = self.get_task_loggers(with_adapters=True)
        data = iter(())
        for logger in loggers.values():
            data = chain(data, logger.get_records(*args, **kwargs))
        return data

    def delete_task_loggers(self):
        """Delete the previous loggers from task logger"""
        loggers = logging.Logger.manager.loggerDict
        for name in list(loggers):
            if name.startswith(self.config.task_logger_basename):
                del loggers[name]

    def clear(self):
        """Clear tasks, parameters etc. of the session"""
        #! TODO: Remove?
        from rocketry.core import Parameters

        self.tasks = set()
        self.parameters = Parameters()

    def __getstate__(self):
        # NOTE: When a process task is executed, it will pickle
        # the task.session. Therefore removing unpicklable here.
        state = self.__dict__.copy()
        state["tasks"] = set()
        state["_cond_cache"] = None
        state["_cond_parsers"] = None
        state["session"] = None
        #state["parameters"] = None
        state['scheduler'] = None
        return state

    @property
    def env(self):
        "Shorthand for parameter 'env'"
        return self.parameters.get("env")

    @env.setter
    def env(self, value):
        "Shorthand for parameter 'env'"
        self.parameters["env"] = value

    def set_as_default(self):
        """Set this session as the default session for
        next tasks, conditions and schedulers that
        are created.
        """

        RedBase.session = self

        import rocketry
        rocketry.session = self

    def hook_startup(self):
        def wrapper(func):
            self.hooks.scheduler_startup.append(func)
            return func
        return wrapper

    def hook_shutdown(self):
        def wrapper(func):
            self.hooks.scheduler_shutdown.append(func)
            return func
        return wrapper

    def hook_scheduler_cycle(self):
        def wrapper(func):
            self.hooks.scheduler_cycle.append(func)
            return func
        return wrapper

    def hook_task_init(self):
        def wrapper(func):
            self.hooks.task_init.append(func)
            return func
        return wrapper

    def hook_task_execute(self):
        def wrapper(func):
            self.hooks.task_execute.append(func)
            return func
        return wrapper



import pytest

from rocketry.time.interval import (
    TimeOfHour,
    TimeOfDay,
)
from rocketry.time import All, Any, always

def test_equal():
    assert not TimeOfHour("10:00") == TimeOfHour("11:00")
    assert TimeOfHour("10:00") == TimeOfHour("10:00")
    assert (TimeOfHour("10:00", "12:00") & TimeOfHour("11:00", "13:00")) == (TimeOfHour("10:00", "12:00") & TimeOfHour("11:00", "13:00"))
    assert (TimeOfHour("10:00", "12:00") | TimeOfHour("11:00", "13:00")) == (TimeOfHour("10:00", "12:00") | TimeOfHour("11:00", "13:00"))

def test_and():
    time = TimeOfHour("10:00", "14:00") & TimeOfHour("09:00", "12:00")
    assert time == All(TimeOfHour("10:00", "14:00"), TimeOfHour("09:00", "12:00"))

def test_and_reduce_always():
    time = always & always
    assert time is always

    time = TimeOfHour("10:00", "14:00") & always
    assert time == TimeOfHour("10:00", "14:00")


def test_any():
    time = TimeOfHour("10:00", "14:00") | TimeOfHour("09:00", "12:00")
    assert time == Any(TimeOfHour("10:00", "14:00"), TimeOfHour("09:00", "12:00"))

def test_any_reduce_always():
    time = always | always
    assert time is always

    time = TimeOfHour("10:00", "14:00") | always
    assert time is always

def test_error():
    with pytest.raises(ValueError):
        Any()
    with pytest.raises(ValueError):
        All()
    with pytest.raises(TypeError):
        Any(1, TimeOfHour())
    with pytest.raises(TypeError):
        All(1, TimeOfHour())

    # Test & and | don't work with non-period
    with pytest.raises(TypeError):
        TimeOfDay() & 1

    with pytest.raises(TypeError):
        TimeOfDay() | 1

def test_str():
    assert str(TimeOfDay("10:00", "12:00") & TimeOfDay("16:00", "17:00")) == '10 hours - 12 hours & 16 hours - 17 hours'
    assert str(TimeOfDay("10:00", "12:00") | TimeOfDay("16:00", "17:00")) == '10 hours - 12 hours | 16 hours - 17 hours'

def test_repr():
    assert repr(TimeOfDay("10:00", "12:00") & TimeOfDay("16:00", "17:00")) == 'All(TimeOfDay(_start=36000000000, _end=43200000000), TimeOfDay(_start=57600000000, _end=61200000000))'
    assert repr(TimeOfDay("10:00", "12:00") | TimeOfDay("16:00", "17:00")) == 'Any(TimeOfDay(_start=36000000000, _end=43200000000), TimeOfDay(_start=57600000000, _end=61200000000))'




import copy
from typing import Callable, Optional, Pattern, Union

from pydantic import Field
from rocketry.args import Session
from rocketry import Session as _Session

from rocketry.core.condition import BaseCondition #, Task
from rocketry.tasks.func import FuncTask


class _FuncTaskCondWrapper(FuncTask):

    # For some reason, the order of cls attrs broke here so we need to reorder then:
    session: _Session
    name: Optional[str] = Field(description="Name of the task. Must be unique")

    def _handle_return(self, value):
        # Handle the return value of the function
        self.session._cond_states[self.name] = value

class TaskCond(BaseCondition):
    """Condition which state is defined by a task

    TaskCond is a similar condition as ``FuncCond``
    except the a task is formed from the function. This
    condition is useful for checks that may be slow
    in terms of IO or by other system resources or could
    get stuck.

    The produced task will run depending on its ``start_cond``
    and the last check is considered to be valid given
    the ``active_time``. For example, you can set the
    task to run every 10 minutes to reduce the amount of
    time the condition needs to be checked.

    .. note::

        Multiple tasks may be created when using a TaskCond
        with different paramters.

    Parameters
    ----------
    func : Callable, optional
        Function to check for the condition state.
        Can also be passed as via decorator.
    active_time : str
        Time how long the previous time check is valid
        until the condition is considered ``False``,
        defaults 'never'
    syntax : str, re.compile
        Syntax for the condition to be used in
        condition creation (ie. Task's start_cond)
    **kwargs : dict
        Passed to ``FuncTask`` as arguments.

    Examples
    --------
    .. code-block:: python

        from rocketry.conditions import TaskCond

        @TaskCond(syntax=re.compile("is foo at (?P<place>.+)"), start_cond="every 10 minutes")
        def is_foo(place):
            ... # Expensive check
            return True

    Using the condition:

    .. code-block:: python

        from rocketry.tasks import FuncTask

        @FuncTask(start_cond="is foo at home")
        def mytask():
            ...

    """

    def __init__(self,
                 session,
                 func: Callable[..., bool]=None,
                 active_time:str ="always",
                 syntax:Union[str, Pattern]=None,
                 **kwargs):
        from rocketry.parse import parse_time

        self.func = func
        self.syntax = syntax
        self.active_time = parse_time(active_time, session=session)

        self.kwds_task = kwargs

        if self.func is not None:
            self._set_parsing()

        self.session = session

    def _set_task(self, *args, **kwargs) -> 'TaskCond':
        "Recreate the condition using args and kwargs"
        new_self = copy.copy(self)

        new_self.task = _FuncTaskCondWrapper(
            func=self.func,
            #on_exists="rename",
            name=f"_condition-{self._get_func_name(self.func)}",
            parameters=kwargs,
            session=self.session,
            **self.kwds_task
        )

        return new_self

    def __call__(self, func: Callable[..., bool]):
        self.func = func
        self._set_parsing()
        return func

    def get_state(self, session=Session()):

        task_state = session._cond_states.get(self.task.name, False)
        if self.task.last_success is None or self.task.last_success not in self.active_time:
            # The cooldown period is gone --> setting to default
            self.state = False
        else:
            self.state = task_state
        return self.state

    def _set_parsing(self):
        from rocketry.parse import CondParser
        self.session._cond_parsers[self.syntax] = CondParser(func=self._set_task, session=self.session, cached=True)

    def _get_func_name(self, func):
        func_module = func.__module__
        func_name = getattr(func, "__name__", type(func).__name__)
        if func_module == "__main__":
            # Showing as 'myfunc'
            return func_name
        else:
            # Showing as 'path.to.module:myfunc'
            return f"{func_module}:{func_name}"



import logging

from rocketry import Rocketry, Grouper
from rocketry.conditions import TaskStarted
from rocketry.conds import daily, time_of_day

def set_logging_defaults():
    task_logger = logging.getLogger("rocketry.task")
    task_logger.handlers = []
    task_logger.setLevel(logging.WARNING)

def test_simple(session):
    set_logging_defaults()

    app = Rocketry(execution="async")

    @app.task(daily)
    def do_things():
        ...

    group = Grouper()

    @group.task(daily)
    def do_other_things():
        ...

    assert app.session.tasks == {app.session["do_things"]}

    app.include_grouper(group)
    assert app.session.tasks == {app.session["do_things"], app.session["do_other_things"]}
    assert app.session["do_other_things"].start_cond == daily

def test_prefix(session):
    set_logging_defaults()

    app = Rocketry(execution="async")

    @app.task(daily)
    def do_things():
        ...

    group = Grouper(prefix="mytests.")

    @group.task(daily)
    def do_things():
        ...

    app.include_grouper(group)
    assert app.session.tasks == {app.session["do_things"], app.session["mytests.do_things"]}

def test_start_cond(session):
    set_logging_defaults()

    app = Rocketry(execution="async")
    group = Grouper(start_cond=time_of_day.between("10:00", "18:00"))

    @group.task(daily)
    def do_things():
        ...

    app.include_grouper(group)
    task = app.session[do_things]
    assert task.start_cond == daily & time_of_day.between("10:00", "18:00")

def test_execution(session):
    set_logging_defaults()

    app = Rocketry(config=dict(task_execution="process"))
    group = Grouper(execution="main")

    @group.task(daily, execution="thread")
    def do_things():
        ...

    @group.task(daily)
    def do_things_2():
        ...

    app.include_grouper(group)
    assert app.session[do_things].execution == "thread"
    assert app.session[do_things_2].execution == "main"

def test_custom_condition(session):
    set_logging_defaults()

    app = Rocketry(execution="async")

    @app.cond()
    def is_foo():
        return True

    group = Grouper()

    @group.cond()
    def is_bar():
        return True

    @group.task(is_foo & is_bar)
    def do_things():
        ...

    app.include_grouper(group)
    assert app.session.tasks == {app.session["do_things"]}

def test_params(session):
    set_logging_defaults()

    app = Rocketry(execution="async")
    app.params(x="hello", z="world")

    group = Grouper()
    group.params(x="hi", y="universe")

    app.include_grouper(group)
    assert dict(app.session.parameters) == {"x": "hi", "y": "universe", "z": "world"}

def test_func_param(session):
    set_logging_defaults()

    app = Rocketry(execution="async")
    @app.param("x")
    def get_x():
        return "hello"

    group = Grouper()
    @group.param("y")
    def get_x():
        return "world"

    app.include_grouper(group)
    assert list(app.session.parameters.keys()) == ["x", "y"]


def test_run(session):
    set_logging_defaults()

    app = Rocketry(config={"task_execution": "main"})

    @app.task(daily)
    def do_things():
        ...

    group = Grouper()

    @group.task(daily)
    def do_other_things():
        ...

    assert app.session.tasks == {app.session["do_things"]}

    app.include_grouper(group)

    app.session.config.shut_cond = TaskStarted(task=do_other_things)
    app.run()
    logger = app.session[do_other_things].logger
    assert logger.filter_by(action="success").count() == 1

def test_run_cond_elsewhere(session):
    set_logging_defaults()

    app = Rocketry(config={"task_execution": "main"})
    group = Grouper()

    @group.cond()
    def is_foo():
        return True

    @app.task(is_foo)
    def do_things():
        ...

    # We don't include the group

    app.session.config.shut_cond = TaskStarted(task=do_things)
    app.run()
    logger = app.session[do_things].logger
    assert logger.filter_by(action="success").count() == 1




"""
Tests for methods if one wish to run
only some parts of the scheduler (like
executing one task)
"""

import asyncio
import time

import pytest

from rocketry.core import Scheduler
from rocketry.tasks import FuncTask
from rocketry.exc import TaskInactionException
from rocketry.conditions import AlwaysFalse

def run_failing():
    raise RuntimeError("Task failed")

def run_succeeding():
    time.sleep(0.05)

def run_inacting():
    raise TaskInactionException()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,run_count,fail_count,success_count",
    [
        pytest.param(
            run_succeeding,
            1, 0, 1,
            id="Succeeding task"),

        pytest.param(
            run_failing,
            1, 1, 0,
            id="Failing task"),
        pytest.param(
            run_inacting,
            1, 0, 0,
            id="Inacting task"),
    ],
)
def test_run_task(execution, task_func, run_count, fail_count, success_count, session):
    "Example of how to run only one task once using the scheduler"

    task = FuncTask(func=task_func, name="task", start_cond=AlwaysFalse(), execution=execution, session=session)
    logger = task.logger

    scheduler = Scheduler(session=session)
    asyncio.run(scheduler.run_task(task))
    assert run_count == logger.filter_by(action="run").count()

    while scheduler.n_alive > 0:
        time.sleep(0.001)
    scheduler.handle_logs()

    assert success_count == logger.filter_by(action="success").count()
    assert fail_count == logger.filter_by(action="fail").count()



import pytest

from rocketry.conditions import (
    TaskFinished,
    TaskFailed,
    TaskSucceeded,

    DependFinish,
    DependFailure,
    DependSuccess,
    TaskStarted,
    TaskTerminated,
)
from rocketry.tasks import FuncTask



def run_task(fail=False):
    print("Running func")
    if fail:
        raise RuntimeError("Task failed")

@pytest.mark.parametrize('execution_number', range(10))
def test_task_status_race(tmpdir, session, execution_number):

    # c = logging.LogRecord("",1, "a", "a", "2", (), "").created
    # c = time.time()
    # t = time.time()
    # c = datetime.datetime.fromtimestamp(c)
    # t = datetime.datetime.now()
    # assert t >= c
    # return

    condition = TaskFinished(task="runned task")
    task = FuncTask(
        run_task,
        name="runned task",
        execution="main",
        session=session
    )
    task(params={"fail": False})

    # Imitating the __bool__
    assert condition.observe(task=task)

@pytest.mark.parametrize(
    "cls,succeeding,expected",
    [
        pytest.param(
            TaskFinished, True,
            True,
            id="TaskFinished Success"),
        pytest.param(
            TaskFinished, False,
            True,
            id="TaskFinished Failure"),

        pytest.param(
            TaskSucceeded, True,
            True,
            id="TaskSucceeded Success"),
        pytest.param(
            TaskSucceeded, False,
            False,
            id="TaskSucceeded Failure"),

        pytest.param(
            TaskFailed, True,
            False,
            id="TaskFailed Success"),
        pytest.param(
            TaskFailed, False,
            True,
            id="TaskFailed Failure"),
    ],
)
def test_task_status(tmpdir, session, cls, succeeding, expected):
    # RACE CONDITION 2021-08-16: 'TaskFailed Failure' failed due to assert bool(condition) if expected else not bool(condition)

    with tmpdir.as_cwd() as old_dir:
        condition = cls(task="runned task")

        task = FuncTask(
            run_task,
            name="runned task",
            execution="main",
            session=session
        )

        # Has not yet ran
        assert not condition.observe(task=task)

        # Now has
        try:
            task(params={"fail": not succeeding})
        except:
            pass

        # we sleep 20ms to
        # There is a very small inaccuracy in time.time() that is used by
        # logging library to create LogRecord.created. On Windows this
        # inaccuracy can be 16ms (https://stackoverflow.com/questions/1938048/high-precision-clock-in-python)
        # and in some cases the task is not registered to have run as
        # in memory logging is way too fast. Therefore we just wait
        # 20ms to fix the issue
        # time.sleep(0.02)
        assert condition.observe(task=task) if expected else not condition.observe(task=task)


@pytest.mark.parametrize(
    "cls,expected",
    [
        pytest.param(
            DependFinish,
            True,
            id="DependFinish"),

        pytest.param(
            DependSuccess,
            False,
            id="DependSuccess"),

        pytest.param(
            DependFailure,
            True,
            id="DependFailure"),
    ],
)
def test_task_depend_fail(tmpdir, session, cls, expected):
    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        condition = cls(task="runned task", depend_task="prerequisite task")

        depend_task = FuncTask(
            run_task,
            name="prerequisite task",
            execution="main",
            session=session
        )

        task = FuncTask(
            run_task,
            name="runned task",
            execution="main",
            session=session
        )

        # ------------------------ t0
        assert not condition.observe(task=task)


        # depend_task
        # -----|------------------- t0
        try:
            depend_task(params={"fail": True})
        except:
            pass
        assert condition.observe(task=task) if expected else not condition.observe(task=task)


        # depend_task     task
        # -----|-----------|------- t0
        task()
        assert not condition.observe(task=task)


        # depend_task     task     depend_task
        # -----|-----------|-----------|----------- t0
        try:
            depend_task(params={"fail": True})
        except:
            pass
        assert condition.observe(task=task) if expected else not condition.observe(task=task)


        # depend_task     task     depend_task     task
        # -----|-----------|-----------|------------|-------- t0
        task()
        assert not condition.observe(task=task)


@pytest.mark.parametrize(
    "cls,expected",
    [
        pytest.param(
            DependFinish,
            True,
            id="DependFinish"),

        pytest.param(
            DependSuccess,
            True,
            id="DependSuccess"),

        pytest.param(
            DependFailure,
            False,
            id="DependFailure"),
    ],
)
def test_task_depend_success(tmpdir, session, cls, expected):
    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        condition = cls(task="runned task", depend_task="prerequisite task")

        depend_task = FuncTask(
            run_task,
            name="prerequisite task",
            execution="main",
            session=session
        )

        task = FuncTask(
            run_task,
            name="runned task",
            execution="main",
            session=session
        )

        # ------------------------ t0
        assert not condition.observe(task=task)


        # depend_task
        # -----|------------------- t0
        depend_task(params={"fail": False})
        assert condition.observe(task=task) if expected else not condition.observe(task=task)


        # depend_task     task
        # -----|-----------|------- t0
        task()
        assert not condition.observe(task=task)


        # depend_task     task     depend_task
        # -----|-----------|-----------|----------- t0
        depend_task(params={"fail": False})
        assert condition.observe(task=task) if expected else not condition.observe(task=task)


        # depend_task     task     depend_task     task
        # -----|-----------|-----------|------------|-------- t0
        task()
        assert not condition.observe(task=task)


@pytest.mark.parametrize(
    "cls,string",
    [
        pytest.param(TaskFinished, "task 'mytask' finished", id="TaskFinished"),
        pytest.param(TaskFailed, "task 'mytask' failed", id="TaskFailed"),
        pytest.param(TaskSucceeded, "task 'mytask' succeeded", id="TaskSucceeded"),
        pytest.param(TaskStarted, "task 'mytask' started", id="TaskStarted"),
        pytest.param(TaskTerminated, "task 'mytask' terminated", id="TaskStarted"),

        pytest.param(DependFinish, "task 'mydep' finished before 'mytask' started", id="DependFinish"),
        pytest.param(DependFailure, "task 'mydep' failed before 'mytask' started", id="DependFailure"),
        pytest.param(DependSuccess, "task 'mydep' succeeded before 'mytask' started", id="DependSuccess"),
    ],
)
def test_display(cls, string, session):
    task = FuncTask(func=lambda: None, name="mytask", session=session)
    depend_task = FuncTask(func=lambda: None, name="mydep", session=session)
    if cls in (DependFinish, DependSuccess, DependFailure):
        s = str(cls(task=task, depend_task=depend_task))
    else:
        s = str(cls(task=task))
    assert s == string



import datetime

from rocketry.time import TimeDelta
from rocketry.core.condition.base import BaseCondition

class IsPeriod(BaseCondition):
    """Condition for checking whether current time
    is in the given time period.

    Parameters
    ----------
    period : rocketry.core.time.TimePeriod
        Time period to check.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("time of day between 10:00 and 15:00")
    IsPeriod(period=TimeOfDay('10:00', '15:00'))

    **Construction example:**

    >>> from rocketry.conditions import IsPeriod
    >>> from rocketry.time import TimeOfDay
    >>> is_morning = IsPeriod(period=TimeOfDay("06:00", "12:00")) # doctest: +SKIP
    """
    def __init__(self, period):
        if isinstance(period, TimeDelta):
            raise AttributeError("TimeDelta does not have __contains__.")
        self.period = period

    def get_state(self):
        return datetime.datetime.now() in self.period

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        elif hasattr(self, "period"):
            return f'currently {str(self.period)}'
        else:
            return type(self).__name__

    def __repr__(self):
        cls_name = type(self).__name__
        return f"{cls_name}(period={repr(self.period)})"



import logging
import pytest

from task_helpers import wait_till_task_finish

from rocketry.tasks import FuncTask
from rocketry.core.task import Task
from rocketry.exc import TaskInactionException, TaskLoggingError
from rocketry.conditions import AlwaysFalse, AlwaysTrue

Task.use_instance_naming = True


def run_successful_func():
    print("Running func")

def run_failing_func():
    print("Running func")
    raise RuntimeError("Task failed")

def run_inaction():
    raise TaskInactionException()

def run_parametrized(integer, string, optional_float=None):
    assert isinstance(integer, int)
    assert isinstance(string, str)
    assert isinstance(optional_float, float)

def run_parametrized_kwargs(**kwargs):
    assert {} != kwargs
    assert isinstance(kwargs["integer"], int)
    assert isinstance(kwargs["string"], str)
    assert isinstance(kwargs["optional_float"], float)

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,expected_outcome,exc_cls",
    [
        pytest.param(
            run_successful_func,
            "success",
            None,
            id="Success"),
        pytest.param(
            run_failing_func,
            "fail",
            RuntimeError,
            id="Failure"),
        pytest.param(
            run_inaction,
            "inaction",
            None,
            id="Inaction"),
    ],
)
def test_run(task_func, expected_outcome, exc_cls, execution, session):
    task = FuncTask(
        task_func,
        name="a task",
        execution=execution,
        session=session
    )

    try:
        task()
    except:
        # failing execution="main"
        if expected_outcome != "fail":
            raise

    # Wait for finish
    wait_till_task_finish(task)

    assert task.status == expected_outcome

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": expected_outcome},
    ] == records


async def run_async_successful():
    ...

async def run_async_fail():
    raise RuntimeError("Failed")

async def run_async_inaction():
    raise TaskInactionException("Did nothing")

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,expected_outcome",
    [
        pytest.param(run_async_successful, "success", id="Success"),
        pytest.param(run_async_fail, "fail", id="Failure"),
        pytest.param(run_async_inaction, "inaction", id="Inaction"),
    ],
)
def test_run_async(task_func, expected_outcome, execution, session):

    task = FuncTask(
        task_func,
        name="a task",
        execution=execution,
        session=session
    )

    try:
        task()
    except:
        # failing execution="main"
        if expected_outcome != "fail":
            raise

    # Wait for finish
    wait_till_task_finish(task)

    assert task.status == expected_outcome

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": expected_outcome},
    ] == records

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,expected_outcome",
    [
        pytest.param(run_async_successful, "success", id="Success"),
        pytest.param(run_async_fail, "fail", id="Failure"),
        pytest.param(run_async_inaction, "inaction", id="Inaction"),
    ],
)
@pytest.mark.asyncio
async def test_run_log_fail_at_start(task_func, expected_outcome, execution, session):
    class MyHandler(logging.Handler):
        def emit(self, record):
            if record.action == "run":
                raise RuntimeError("Oops")
    logger = logging.getLogger("rocketry.task")
    logger.handlers.insert(0, MyHandler())

    task = FuncTask(
        task_func,
        name="a task",
        execution=execution,
        session=session
    )
    if execution != "thread":
        with pytest.raises(TaskLoggingError):
            await task.start_async()
    else:
        await task.start_async()
        assert task._thread_error
    # Wait for finish
    await session.scheduler.wait_task_alive()
    session.scheduler.handle_logs()

    # At the moment the run log is not propagated to
    # finish
    assert task.status != "run"

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,expected_outcome",
    [
        pytest.param(run_async_successful, "success", id="Success"),
        pytest.param(run_async_fail, "fail", id="Failure"),
        pytest.param(run_async_inaction, "inaction", id="Inaction"),
    ],
)
@pytest.mark.asyncio
async def test_run_log_fail_at_end(task_func, expected_outcome, execution, session):
    class MyHandler(logging.Handler):
        def emit(self, record):
            if record.action != "run":
                raise RuntimeError("Oops")
    logger = logging.getLogger("rocketry.task")
    logger.handlers.insert(0, MyHandler())

    task = FuncTask(
        task_func,
        name="a task",
        execution=execution,
        session=session
    )
    if execution == "main":
        with pytest.raises(TaskLoggingError):
            await task.start_async()
    else:
        await task.start_async()

    await session.scheduler.wait_task_alive()

    if execution == "process":
        with pytest.raises(TaskLoggingError):
            session.scheduler.handle_logs()
    assert task.status == "fail"

def test_force_run(session):

    task = FuncTask(
        run_successful_func,
        name="task",
        start_cond=AlwaysFalse(),
        execution="main",
        session=session
    )
    with pytest.deprecated_call():
        task.force_run = True

    assert bool(task)
    assert bool(task)

    task()
    assert not task.force_run


def test_dependency(tmpdir, session):

    task_a = FuncTask(
        run_successful_func,
        name="task_a",
        start_cond=AlwaysTrue(),
        execution="main",
        session=session
    )
    task_b = FuncTask(
        run_successful_func,
        name="task_b",
        start_cond=AlwaysTrue(),
        execution="main",
        session=session
    )
    task_dependent = FuncTask(
        run_successful_func,
        name="task_dependent",
        start_cond="after task 'task_a' & after task 'task_b'",
        execution="main",
        session=session
    )
    assert not bool(task_dependent)
    task_a()
    assert not bool(task_dependent)
    task_b()
    assert bool(task_dependent)


# Parametrization
def test_parametrization_runtime(session):

    task = FuncTask(
        run_parametrized,
        name="a task",
        execution="main",
        session=session
    )

    task(params={"integer": 1, "string": "X", "optional_float": 1.1, "extra_parameter": "Should not be passed"})

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": "success"},
    ] == records

def test_parametrization_local(session):

    task = FuncTask(
        run_parametrized,
        name="a task",
        parameters={"integer": 1, "string": "X", "optional_float": 1.1},
        execution="main",
        session=session
    )

    task()

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": "success"},
    ] == records

def test_parametrization_kwargs(session):

    task = FuncTask(
        run_parametrized_kwargs,
        name="a task",
        parameters={"integer": 1, "string": "X", "optional_float": 1.1},
        execution="main",
        session=session
    )

    task()

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": "success"},
    ] == records



import re
from functools import partial

from rocketry.conditions.task import *
from rocketry.conditions.scheduler import *
from rocketry.conditions.time import *
from rocketry.conditions.parameter import ParamExists, IsEnv

from rocketry.session import Session
from rocketry.core.condition import Not, BaseCondition

from rocketry.conds import (
    secondly, minutely, hourly, daily, weekly, monthly, every,
    cron
)

def _from_period_task_has(cls, span_type=None, inverse=False, **kwargs):
    from rocketry.time.construct import get_full_cycle, get_between, get_after, get_before
    from rocketry.time import TimeDelta

    period_func = {
        "between": get_between,
        "after": get_after,
        "before": get_before,
        "starting": get_full_cycle,
        None: get_full_cycle,
        "every": TimeDelta,
        "on": get_on,

        "past": TimeDelta,
    }[span_type]

    task = kwargs.pop("task", None)
    period = period_func(**kwargs)

    cls_kwargs = {"task": task} if task is not None else {}
    if inverse:
        return Not(cls(period=period, **cls_kwargs))
    else:
        return cls(period=period, **cls_kwargs)


def _set_is_period_parsing():

    from functools import partial

    def _get_is_period(period_constructor, *args, **kwargs):
        period = period_constructor(*args, **kwargs)
        return IsPeriod(period=period)

    cond_parsers = Session._cls_cond_parsers
    time_parsers = Session._time_parsers

    cond_parsers.update(
        {
            parsing: partial(_get_is_period, period_constructor=parser)
            for parsing, parser in time_parsers.items()
        }
    )

def _set_task_has_parsing():

    cond_parsers = Session._cls_cond_parsers

    clss = [
        ("failed", TaskFailed),
        ("succeeded", TaskSucceeded),
        ("finished", TaskFinished),
        ("terminated", TaskTerminated),
        ("inacted", TaskInacted),
        ("started", TaskStarted)
    ]
    for (action, cls) in clss:
        func = partial(_from_period_task_has, cls=cls)
        for prefix in ("", r"task '(?P<task>.+)' "):
            cond_parsers.update(
                {
                    re.compile(fr"{prefix}has {action}"): cls,
                    re.compile(fr"{prefix}has {action} (?P<type_>this month|this week|today|this hour|this minute) (?P<span_type>starting) (?P<start>.+)"): func,
                    re.compile(fr"{prefix}has {action} (?P<type_>this month|this week|today|this hour|this minute) (?P<span_type>between) (?P<start>.+) and (?P<end>.+)"): func,
                    re.compile(fr"{prefix}has {action} (?P<type_>this month|this week|today|this hour|this minute) (?P<span_type>after) (?P<start>.+)"): func,
                    re.compile(fr"{prefix}has {action} (?P<type_>this month|this week|today|this hour|this minute) (?P<span_type>before) (?P<end>.+)"): func,
                    re.compile(fr"{prefix}has {action} (?P<type_>this month|this week|today|this hour|this minute)"): func,
                    re.compile(fr"{prefix}has {action} (?P<type_>this month|this week|today|this hour|this minute) (?P<span_type>on) (?P<start>.+)"): func,
                    re.compile(fr"{prefix}has {action} (in )?past (?P<past>.+)"): partial(func, span_type='past'),
                }
            )

def _set_scheduler_parsing():

    cond_parsers = Session._cls_cond_parsers

    cls = SchedulerStarted
    func = partial(_from_period_task_has, cls=cls)
    cond_parsers.update(
        {
            re.compile(fr"scheduler has run over (?P<past>.+)"): partial(func, span_type='past', inverse=True),
            re.compile(fr"scheduler started (?P<past>.+) ago"): partial(func, span_type='past'),

            re.compile(r"scheduler has more than (?P<__gt__>[0-9]+) cycles"): SchedulerCycles.from_magic,
            re.compile(r"scheduler has less than (?P<__lt__>[0-9]+) cycles"): SchedulerCycles.from_magic,
            re.compile(r"scheduler has (?P<__eq__>[0-9]+) cycles"): SchedulerCycles.from_magic,
        }
    )

def _set_task_exec_parsing():
    cond_parsers = Session._cls_cond_parsers

    conds = {
        "secondly": secondly,
        "minutely": minutely,
        "hourly": hourly,
        "daily": daily,
        "weekly": weekly,
        "monthly": monthly,
    }
    options = {
        r' before (?P<end>.+)': 'before',
        r' between (?P<start>.+) and (?P<end>.+)': 'between',
        r' after (?P<start>.+)': 'after',
        r' starting (?P<start>.+)': 'starting',
        r' on (?P<span>.+)': 'on',
    }

    for str_period, cond in conds.items():
        cond_parsers[str_period] = cond

        for str_option, method_name in options.items():
            syntax = f"{str_period}{str_option}"
            method = getattr(cond, method_name)

            # Add to the syntax
            cond_parsers[re.compile(syntax)] = method
    # Add "every ..."
    cond_parsers[re.compile(r"every (?P<past>.+)")] = every

    # Cron
    cond_parsers[re.compile(r"cron (?P<__expr>.+)")] = cron

def _set_task_running_parsing():
    cond_parsers = Session._cls_cond_parsers

    cond_parsers.update(
        {
            re.compile(r"while task '(?P<task>.+)' is running"): TaskRunning,
            re.compile(r"task '(?P<task>.+)' is running"): TaskRunning,
        }
    )

def _set_task_pipelining_parsing():
    cond_parsers = Session._cls_cond_parsers

    cond_parsers.update(
        {
            re.compile(r"after task '(?P<depend_task>.+)'( succeeded)?"): DependSuccess,
            re.compile(r"after tasks '(?P<depend_tasks>.+)'( succeeded)?"): DependSuccess._parse_multi_all,
            re.compile(r"after any tasks '(?P<depend_tasks>.+)'( succeeded)?"): DependSuccess._parse_multi_any,

            re.compile(r"after task '(?P<depend_task>.+)' failed"): DependFailure,
            re.compile(r"after tasks '(?P<depend_tasks>.+)' failed"): DependFailure._parse_multi_all,
            re.compile(r"after any tasks '(?P<depend_tasks>.+)' failed"): DependFailure._parse_multi_any,

            re.compile(r"after task '(?P<depend_task>.+)' finished"): DependFinish,
            re.compile(r"after tasks '(?P<depend_tasks>.+)' finished"): DependFinish._parse_multi_all,
            re.compile(r"after any tasks '(?P<depend_tasks>.+)' finished"): DependFinish._parse_multi_any,
        }
    )

def _set_misc_parsing():
    cond_parsers = Session._cls_cond_parsers

    cond_parsers.update(
        {
            re.compile(r"env '(?P<env>.+)'"): IsEnv,
            re.compile(r"param '(?P<l>.+)' exists"): ParamExists._from_list,
            re.compile(r"param '(?P<key>.+)' is '(?P<value>.+)'"): ParamExists._from_key_value,
        }
    )

_set_is_period_parsing()
_set_task_has_parsing()
_set_scheduler_parsing()
_set_task_exec_parsing()

_set_task_running_parsing()
_set_task_pipelining_parsing()
_set_misc_parsing()



import pickle
from inspect import isfunction

import pytest

from rocketry.tasks import FuncTask
from rocketry.conditions import TaskFailed


def func_on_main_level():
    pass

def pickle_dump_read(obj):
    p = pickle.dumps(obj)
    return pickle.loads(p)

class TestFunc:

    def test_func_on_main(self, session):
        task = FuncTask(func_on_main_level, session=session)
        pick_task = pickle_dump_read(task)
        assert pick_task.func.__name__ == "func_on_main_level"
        assert isfunction(pick_task.func)

    def test_func_nested(self, session):
        # This cannot be pickled (cannot use execution == process)
        def func_nested():
            pass
        task = FuncTask(func_nested, execution="process", name="unpicklable", session=session)
        with pytest.raises(AttributeError):
            pickle_dump_read(task)
        # This should not raise (though still not pickleable)
        task = FuncTask(func_nested, execution="thread", name="picklable", session=session)

    def test_unpicklable_start_cond(self, session):
        def func_nested():
            pass
        unpkl_task = FuncTask(func_nested, execution="thread", session=session)
        task = FuncTask(func_on_main_level, execution="process", start_cond=TaskFailed(task=unpkl_task), session=session)

        pick_task = pickle_dump_read(task)
        assert pick_task.func.__name__ == "func_on_main_level"
        assert isfunction(pick_task.func)
        assert pick_task.start_cond is None

    def test_unpicklable_end_cond(self, session):
        def func_nested():
            pass
        unpkl_task = FuncTask(func_nested, execution="thread", session=session)
        task = FuncTask(func_on_main_level, execution="process", end_cond=TaskFailed(task=unpkl_task), session=session)

        pick_task = pickle_dump_read(task)
        assert pick_task.func.__name__ == "func_on_main_level"
        assert isfunction(pick_task.func)
        assert pick_task.end_cond is None

    def test_unpicklable_session(self, session):
        def func_nested():
            pass
        unpkl_task = FuncTask(func_nested, execution="thread", name="unpicklable", session=session)
        task = FuncTask(func_on_main_level, execution="process", name="picklable", session=session)

        assert session.tasks == {unpkl_task, task}

        pick_task = pickle_dump_read(task)
        assert pick_task.func.__name__ == "func_on_main_level"
        assert isfunction(pick_task.func)
        assert pick_task.session is not session
        pickle_dump_read(pick_task.session)

    def test_unpicklable_session_params(self, session):
        session.parameters["unpicklable"] = FuncTask(lambda:None, execution="main", name="unpicklable", session=session)
        session.parameters["picklable"] = "myval"
        task = FuncTask(func_on_main_level, execution="process", name="picklable", session=session)
        pick_task = pickle_dump_read(task)

        assert pick_task.session is None



from typing import Callable
from dataclasses import dataclass

from rocketry.core.time.base import TimePeriod, always

from .interval import TimeOfHour, TimeOfDay, TimeOfMinute, TimeOfWeek, TimeOfMonth, TimeOfYear

@dataclass(frozen=True)
class Cron(TimePeriod):

    minute: str = "*"
    hour: str = "*"
    day_of_month: str = "*"
    month: str = "*"
    day_of_week: str = "*"

    def __init__(self, minute="*", hour="*", day_of_month="*", month="*", day_of_week="*"):
        object.__setattr__(self, "minute", minute)
        object.__setattr__(self, "hour", hour)
        object.__setattr__(self, "day_of_month", day_of_month)
        object.__setattr__(self, "month", month)
        object.__setattr__(self, "day_of_week", day_of_week)

        # *: any value
        # ,: list of values
        # -: range of values
        # /: step values

    def rollforward(self, dt):
        "Get previous time interval of the period."
        return self.get_subperiod().rollforward(dt)

    def rollback(self, dt):
        "Get previous time interval of the period."
        return self.get_subperiod().rollback(dt)

    def _get_period_from_expr(self, cls, expression:str, conv:Callable=None, default=always):

        conv = (lambda i: i) if conv is None else conv
        exprs = expression.split(",")
        full_period = None
        for expr in exprs:
            if expr == "*":
                # Any
                continue
            if "/" in expr:
                expr, step = expr.split("/")
                step_period = cls.create_range(step=int(step))
            else:
                step = None

            if "-" in expr:
                # From to
                start, end = expr.split("-")
                if start.isdigit():
                    start = conv(int(start))
                if end.isdigit():
                    end = conv(int(end))

                if step is not None:
                    # Unlike in traditional Python ranges,
                    # cron includes also the endpoint thus
                    # we convert to int (if needed) and add
                    # one
                    if isinstance(end, str):
                        end = cls._unit_mapping[end.lower()]
                    end += 1
                    period = cls.create_range(start, end, step=int(step))
                else:
                    period = cls(start, end)
            else:
                # At
                step_period = cls.create_range(step=int(step)) if step is not None else always
                value = conv(int(expr)) if expr.isdigit() else expr
                if value == "*":
                    period = always & step_period
                else:
                    period = cls.at(value) & step_period

            if full_period is None:
                full_period = period
            else:
                full_period |= period

        if full_period is None:
            return default
        return full_period

    def _convert_day_of_week(self, i:int):
        # In cron, 0 is Sunday but DayOfWeek don't allow zero
        return 7 if i == 0 else i

    def get_subperiod(self):
        minutely = TimeOfMinute()

        day_of_month = self._get_period_from_expr(TimeOfMonth, self.day_of_month)
        day_of_week = self._get_period_from_expr(TimeOfWeek, self.day_of_week, conv=self._convert_day_of_week)

        if day_of_month is not always and day_of_week is not always:
            # Both specified: OR
            day_of_week_month = day_of_month | day_of_week
        else:
            # Either is specified or neither: AND (the other should be "always")
            day_of_week_month = day_of_month & day_of_week

        return (
            (self._get_period_from_expr(TimeOfHour, self.minute) & minutely)
            & self._get_period_from_expr(TimeOfDay, self.hour)
            & self._get_period_from_expr(TimeOfYear, self.month)
            & day_of_week_month
        )



from rocketry.time import TimeDelta
from rocketry.conditions import SchedulerCycles, SchedulerStarted

def test_scheduler_started(session):

    session.config.shut_cond = ~SchedulerStarted(period=TimeDelta("1 second"))
    session.start()

    assert session.scheduler.n_cycles > 1

def test_scheduler_cycles(session):

    session.config.shut_cond = SchedulerCycles() >= 4
    session.start()

    assert session.scheduler.n_cycles == 4




import datetime
import logging
import warnings

import pytest

from rocketry import Session
from rocketry.session import Config
from rocketry.core import Task, Scheduler, BaseCondition, BaseArgument, Parameters
from rocketry.conds import true

def assert_default(session:Session):
    for cls in (Task, Scheduler, BaseCondition, BaseArgument, Parameters):
        assert cls.session is session


def test_empty():
    session = Session(config={'task_execution': 'async'})
    session.set_as_default()
    assert session.parameters.to_dict() == {}
    assert session.returns.to_dict() == {}
    assert session.tasks == set()

    config = session.config

    assert isinstance(config, Config)
    assert not config.silence_task_prerun
    assert not config.silence_task_logging
    assert not config.silence_cond_check
    assert config.task_pre_exist == 'raise'
    assert config.timeout == datetime.timedelta(minutes=30)
    #assert config.task_execution == 'main'

    assert session.env is None
    assert_default(session)

def test_timeout_parse():
    session = Session(config={"timeout": 0.1, 'task_execution': 'async'})
    assert session.config.timeout == datetime.timedelta(seconds=0.1)

    session = Session(config={"timeout": "0.1 seconds", 'task_execution': 'async'})
    assert session.config.timeout == datetime.timedelta(seconds=0.1)

    session = Session(config={"timeout": datetime.timedelta(seconds=0.1), 'task_execution': 'async'})
    assert session.config.timeout == datetime.timedelta(seconds=0.1)

def test_create():
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        s = Session(config=Config(task_priority=10, task_execution="async"))
        assert s.config.task_priority == 10

        s = Session(config=dict(task_priority=10, task_execution= 'async'))
        assert s.config.task_priority == 10

        Session(config={'task_execution': 'async'})

        s = Session(parameters={"x": 5}, config={'task_execution': 'async'})
        assert s.parameters == Parameters({"x": 5})
        s = Session(parameters=Parameters({"x": 5}), config={'task_execution': 'async'})
        assert s.parameters == Parameters({"x": 5})

    with pytest.warns(FutureWarning):
        Session(config=None)

    with pytest.warns(FutureWarning):
        Session()

    with pytest.raises(TypeError):
        Session(config="invalid")

def test_logging_level():
    task_logger = logging.getLogger("rocketry.task")
    task_logger.setLevel(logging.INFO)
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        # Should not raise warning
        s = Session(config={'task_execution': 'async'})
        s.config.shut_cond = true
        s.start()

    task_logger.setLevel(logging.DEBUG)
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        # Should not raise warning
        s = Session(config={'task_execution': 'async'})
        s.config.shut_cond = true
        s.start()

    task_logger.setLevel(logging.WARNING)
    with pytest.warns(UserWarning):
        s = Session(config={'task_execution': 'async'})
        s.config.shut_cond = true
        s.start()

    # Level is changed to INFO
    assert task_logger.getEffectiveLevel() == logging.INFO



class ClosureParser:
    """

    Manipulate closures, ie. bracets/parentheses/etc.

    Examples:
    ---------
        parser = ClosureParser("(", ")")
        parser.reduce("x + ( 5 * (x / y) + 5)", lambda c: c[1:-2])
        >>> "x +  5 * x / y + 5"

    """
    # TODO: Test

    # TODO: get offsets (closure len) using match for better regex support
    # TODO: check if next_closure is opening/closing using regex
    # TODO: .apply(string, func) : Visit all closures and apply the function

    def __init__(self, opening="(", closing=")"):
        self.opening = opening
        self.closing = closing

    def to_list(self, string):
        "Turn the string to (nested) list of strings of closures"
        # "1 * (21 + (211 - 212) / 22) ^ 3 * (21 + (211 - 212) / 22)"
        # --> ("1 * ", ("21 + ", ("211 - 212",), " / 22"), " ^ 3")
        res = []

        len_open = len(self.opening)
        len_close = len(self.closing)
        while self.count(string):

            start, end = self.find_outer_indices(string)

            # Add left residual
            if string[:start]:
                res.append(string[:start])

            substr = string[slice(start + len_open, end)]
            subres = self.to_list(substr)
            res.append(subres)
            string = string[end+len_close:]

        # Add residual
        if string:
            res.append(string)
        return res

    def find_outer_indices(self, string, start=None):
        "Find indices of next outer closure (ie. '(... (...) ...)')"
        # Start of the fetch (searches from left to right)
        offset = len(self.opening)
        start = 0 if start is None else start
        start -= offset

        openings = 0
        closings = 0
        while openings != closings or openings == 0:

            index, next_closure = self._get_next_element(string, start + offset)

            if next_closure == self.opening:
                openings += 1
                if openings == 1:
                    opening = index
                offset = len(self.opening)
            elif next_closure == self.closing:
                closings += 1
                closing = index
                offset = len(self.closing)

            start = index

        return (opening, closing)

    def count(self, string):
        "Count number of closures in the string"
        return string.count(self.opening)

    def _get_next_element(self, string, start=None, end=None):
        closures = (self.opening, self.closing)
        indexes = {
            string.index(substr, start, end): substr
            for substr in closures
            if substr in string[start:end]
        }
        minimum = min(indexes)
        return minimum, indexes[minimum]



import datetime
import pytest
from rocketry.time import Cron
from rocketry.time.interval import TimeOfDay, TimeOfHour, TimeOfMinute, TimeOfMonth, TimeOfWeek, TimeOfYear

every_minute = TimeOfMinute()

@pytest.mark.parametrize(
    "period,expected",
    [
        # Test at
        pytest.param(Cron(), every_minute, id="* * * * *"),
        pytest.param(Cron("30", "*", "*", "*", "*"), every_minute & TimeOfHour.at(30), id="30 * * * *"),
        pytest.param(Cron("*", "12", "*", "*", "*"), every_minute & TimeOfDay.at(12), id="* 12 * * *"),
        pytest.param(Cron("*", "*", "28", "*", "*"), every_minute & TimeOfMonth.at(28), id="* * 28 * *"),
        pytest.param(Cron("*", "*", "*", "6", "*"), every_minute & TimeOfYear.at(6), id="* * * 6 *"),
        pytest.param(Cron("*", "*", "*", "*", "0"), every_minute & TimeOfWeek.at("Sunday"), id="* * * * 0"),

        # Test at synonyms
        pytest.param(Cron("*", "*", "*", "JUN", "*"), every_minute & TimeOfYear.at("June"), id="* * * JUN *"),
        pytest.param(Cron("*", "*", "*", "*", "SUN"), every_minute & TimeOfWeek.at("Sunday"), id="* * * * SUN"),

        # Test ranges
        pytest.param(Cron("45-59", "*", "*", "*", "*"), every_minute & TimeOfHour(45, 59), id="45-59 * * * *"),
        pytest.param(Cron("*", "10-13", "*", "*", "*"), every_minute & TimeOfDay(10, 13), id="* 10-13 * * *"),
        pytest.param(Cron("*", "*", "28-30", "*", "*"), every_minute & TimeOfMonth(28, 30), id="* * 28-30 * *"),
        pytest.param(Cron("*", "*", "*", "FEB-MAR", "*"), every_minute & TimeOfYear("feb", "mar"), id="* * * FEB-MAR *"),
        pytest.param(Cron("*", "*", "*", "*", "FRI-SUN"), every_minute & TimeOfWeek("fri", "sun"), id="* * * * FRI-SUN"),

        # Test list
        pytest.param(Cron("0,15,30,45", "*", "*", "*", "*"), every_minute & (TimeOfHour.at(0) | TimeOfHour.at(15) | TimeOfHour.at(30) | TimeOfHour.at(45)), id="0,15,30,45 * * * *"),

        # Test combinations
        pytest.param(
            Cron("45-59", "10-13", "28-30", "FEB-MAR", "FRI-SUN"),
            every_minute & TimeOfHour(45, 59) & TimeOfDay(10, 13) & TimeOfYear("feb", "mar") & (TimeOfMonth(28, 30) | TimeOfWeek("fri", "sun")),
            id="45-59 10-13 28-30 FEB-MAR FRI-SUN"
        ),
        pytest.param(
            Cron("45-59", "10-13", "28-30", "FEB-MAR", "*"),
            every_minute & TimeOfHour(45, 59) & TimeOfDay(10, 13) & TimeOfYear("feb", "mar") & TimeOfMonth(28, 30),
            id="45-59 10-13 28-30 FEB-MAR *"
        ),
        pytest.param(
            Cron("0-29,45-59", "0-10,20-23", "*", "*", "*"),
            (TimeOfHour(0, 29) | TimeOfHour(45, 59)) & (TimeOfDay(0, 10) | TimeOfDay(20, 23)) & every_minute,
            id="0-29,45-59 0-10,20-23 * * *"
        ),

        # Test skip
        pytest.param(
            Cron("*/15", "*", "*", "*", "*"),
            every_minute & (
                TimeOfHour.at("00:00") | TimeOfHour.at("15:00") | TimeOfHour.at("30:00") | TimeOfHour.at("45:00")
            ),
            id="*/15 * * * *"
        ),
        pytest.param(
            Cron("*", "*/6", "*", "*", "*"),
            every_minute & (
                TimeOfDay.at("00:00") | TimeOfDay.at("06:00") | TimeOfDay.at("12:00") | TimeOfDay.at("18:00")
            ),
            id="* */6 * * *"
        ),
        pytest.param(
            Cron("*", "2-17/6", "*", "*", "*"),
            every_minute & (
                TimeOfDay.at("02:00") | TimeOfDay.at("08:00") | TimeOfDay.at("14:00")
            ),
            id="* 2-17/6 * * *"
        ),
        pytest.param(
            Cron("*", "*", "*", "*", "Tue-Fri/2"),
            every_minute & (
                TimeOfWeek.at("Tue") | TimeOfWeek.at("Thu")
            ),
            id="* * * * Tue-Fri/2"
        ),
        pytest.param(
            Cron("*", "*", "*", "Feb-Aug/3", "*"),
            every_minute & (
                TimeOfYear.at("Feb") | TimeOfYear.at("May") | TimeOfYear.at("Aug")
            ),
            id="* * * Feb-Aug/3 *"
        ),
    ]
)
def test_subperiod(period, expected):
    subperiod = period.get_subperiod()
    assert subperiod == expected

def test_in():
    period = Cron("30", "*", "*", "*", "*")
    assert datetime.datetime(2022, 8, 7, 12, 29, 59) not in period
    assert datetime.datetime(2022, 8, 7, 12, 30, 00) in period
    assert datetime.datetime(2022, 8, 7, 12, 30, 59) in period
    assert datetime.datetime(2022, 8, 7, 12, 31, 00) not in period

def test_roll_forward_simple():
    period = Cron("30", "*", "*", "*", "*")

    # Roll tiny amount
    interv = period.rollforward(datetime.datetime(2022, 8, 7, 12, 29, 59))
    assert interv.closed == 'left'
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 00)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 31, 00)

    # No roll (at left)
    interv = period.rollforward(datetime.datetime(2022, 8, 7, 12, 30, 0))
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 00)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 31, 00)

    # No roll (at center)
    interv = period.rollforward(datetime.datetime(2022, 8, 7, 12, 30, 30))
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 30)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 31, 00)
    assert interv.closed == 'left'

    # No roll (at right)
    interv = period.rollforward(datetime.datetime(2022, 8, 7, 12, 30, 59, 999999))
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 59, 999999)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 31, 00)
    assert interv.closed == 'left'

    # Roll (at right)
    interv = period.rollforward(datetime.datetime(2022, 8, 7, 12, 31))
    assert interv.left == datetime.datetime(2022, 8, 7, 13, 30)
    assert interv.right == datetime.datetime(2022, 8, 7, 13, 31)
    assert interv.closed == 'left'

def test_roll_back_simple():
    period = Cron("30", "*", "*", "*", "*")

    # Roll tiny amount
    interv = period.rollback(datetime.datetime(2022, 8, 7, 12, 31, 1))
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 00)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 31, 00)
    assert interv.closed == "left"

    # No roll (at left, single point)
    interv = period.rollback(datetime.datetime(2022, 8, 7, 12, 30, 0))
    assert interv.left == datetime.datetime(2022, 8, 7, 11, 30, 00)
    assert interv.right == datetime.datetime(2022, 8, 7, 11, 31, 00)
    assert interv.closed == "left"

    # No roll (at center)
    interv = period.rollback(datetime.datetime(2022, 8, 7, 12, 30, 30))
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 00)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 30, 30)
    assert interv.closed == "left"

    # No roll (at right)
    interv = period.rollback(datetime.datetime(2022, 8, 7, 12, 30, 59, 999999))
    assert interv.left == datetime.datetime(2022, 8, 7, 12, 30, 00)
    assert interv.right == datetime.datetime(2022, 8, 7, 12, 30, 59, 999999)

    # Roll (at right)
    interv = period.rollback(datetime.datetime(2022, 8, 7, 14, 15))
    assert interv.left == datetime.datetime(2022, 8, 7, 13, 30)
    assert interv.right == datetime.datetime(2022, 8, 7, 13, 31)

def test_roll_minute_range():
    period = Cron("30-45", "*", "*", "*", "*")

    interv = period.rollforward(datetime.datetime.fromisoformat("2022-08-07 12:33:00"))
    assert interv.left == datetime.datetime.fromisoformat("2022-08-07 12:33:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-08-07 12:34:00")

    interv = period.rollback(datetime.datetime.fromisoformat("2022-08-07 12:32:59"))
    assert interv.left == datetime.datetime.fromisoformat("2022-08-07 12:32:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-08-07 12:32:59")

def test_roll_complex():
    period = Cron(*"15,30 18-22 20 OCT *".split(" "))

    interv = period.rollforward(datetime.datetime(2022, 8, 7, 10, 0, 0))
    assert interv.left == datetime.datetime.fromisoformat("2022-10-20 18:15:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-10-20 18:16:00")

    interv = period.rollback(datetime.datetime(2022, 12, 7, 10, 0, 0))
    assert interv.left == datetime.datetime.fromisoformat("2022-10-20 22:30:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-10-20 22:31:00")

def test_roll_conflict_day_of_week_first():
    # If day_of_month and day_of_week are passed
    # Crontab seems to prefer the one that is sooner (OR)
    period = Cron(*"15 18-22 20 OCT MON".split(" "))

    # Prefer day of week
    interv = period.rollforward(datetime.datetime(2022, 8, 7, 10, 0, 0))
    assert interv.left == datetime.datetime.fromisoformat("2022-10-03 18:15:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-10-03 18:16:00")

    # Prefer day of week
    interv = period.rollback(datetime.datetime(2022, 12, 7, 10, 0, 0))
    assert interv.left == datetime.datetime.fromisoformat("2022-10-31 22:15:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-10-31 22:16:00")

def test_roll_conflict_day_of_month_first():
    # If day_of_month and day_of_week are passed
    # Crontab seems to prefer the one that is sooner (OR)
    period = Cron(*"15 18-22 3 OCT FRI".split(" "))

    interv = period.rollforward(datetime.datetime(2022, 8, 7, 10, 0, 0))
    assert interv.left == datetime.datetime.fromisoformat("2022-10-03 18:15:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-10-03 18:16:00")

    period = Cron(*"15 18-22 29 OCT FRI".split(" "))
    interv = period.rollback(datetime.datetime(2022, 12, 7, 10, 0, 0))
    assert interv.left == datetime.datetime.fromisoformat("2022-10-29 22:15:00")
    assert interv.right == datetime.datetime.fromisoformat("2022-10-29 22:16:00")



from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo
from .log_record import MinimalRecord

def create_default_handler():
    "Create default handler that can be read"
    return RepoHandler(
        repo=MemoryRepo(model=MinimalRecord)
    )



from rocketry.conditions import (
    DependFailure, DependFinish, DependSuccess, TaskFailed,
    TaskFinished, TaskRunnable,
    TaskStarted, TaskSucceeded,
    Retry,
    SchedulerStarted, SchedulerCycles,
)
from rocketry.core import (
    BaseCondition
)
from rocketry.core.condition import (
    AlwaysTrue, AlwaysFalse,
)
from rocketry.core.condition.base import All, Any
from rocketry.time import Cron

from rocketry.time import (
    TimeOfSecond, TimeOfMinute, TimeOfHour,
    TimeOfDay, TimeOfWeek, TimeOfMonth,
    TimeDelta, TimeSpanDelta
)

from .time import IsPeriod
from .task import TaskExecutable, TaskRunning

# Utility classes
# ---------------

class TimeCondWrapper(BaseCondition):

    def __init__(self, cls_cond, cls_period, **kwargs):
        self._cls_cond = cls_cond
        self._cls_period = cls_period
        self._cond_kwargs = kwargs

    def between(self, start, end):
        period = self._cls_period(start, end)
        return self._get_cond(period)

    def before(self, end):
        period = self._cls_period(None, end)
        return self._get_cond(period)

    def after(self, start):
        period = self._cls_period(start, None)
        return self._get_cond(period)

    def on(self, span):
        # Alias for "at"
        return self.at(span)

    def at(self, span):
        # Alias for "on"
        period = self._cls_period(span, time_point=True)
        return self._get_cond(period)

    def starting(self, start):
        period = self._cls_period.starting(start)
        return self._get_cond(period)

    def observe(self, **kwargs):
        cond = self.get_cond()
        return cond.observe(**kwargs)

    def get_cond(self):
        "Get condition the wrapper itself represents"
        period = self._cls_period(None, None)
        return self._get_cond(period)

    def _get_cond(self, period):
        return self._cls_cond(period=period, **self._cond_kwargs)

class TimeActionWrapper(BaseCondition):

    def __init__(self, cls_cond, task=None):
        self.cls_cond = cls_cond
        self.task = task

    def observe(self, **kwargs):
        cond = self.get_cond()
        return cond.observe(**kwargs)

    def __call__(self, task):
        return TimeActionWrapper(self.cls_cond, task=task)

    @property
    def this_minute(self):
        return self._get_wrapper(TimeOfMinute)

    @property
    def this_hour(self):
        return self._get_wrapper(TimeOfHour)

    @property
    def this_day(self):
        return self._get_wrapper(TimeOfDay)

    @property
    def this_week(self):
        return self._get_wrapper(TimeOfWeek)

    @property
    def this_month(self):
        return self._get_wrapper(TimeOfMonth)

    @property
    def today(self):
        return self.this_day

    def _get_wrapper(self, cls_period):
        return TimeCondWrapper(cls_cond=self.cls_cond, cls_period=cls_period, task=self.task)

    def get_cond(self):
        "Get condition the wrapper represents"
        return self.cls_cond(task=self.task)

class RetryWrapper(BaseCondition):

    def __call__(self, n:int):
        return Retry(n)

    def observe(self, **kwargs):
        return self.get_cond().observe(**kwargs)

    def get_cond(self):
        "Get condition the wrapper represents"
        return Retry(-1)

# Basics
# ------

true = AlwaysTrue()
false = AlwaysFalse()

# Execution related
# -----------------

secondly = TimeCondWrapper(TaskExecutable, TimeOfSecond)
minutely = TimeCondWrapper(TaskExecutable, TimeOfMinute)
hourly = TimeCondWrapper(TaskExecutable, TimeOfHour)
daily = TimeCondWrapper(TaskExecutable, TimeOfDay)
weekly = TimeCondWrapper(TaskExecutable, TimeOfWeek)
monthly = TimeCondWrapper(TaskExecutable, TimeOfMonth)

# Time related
# ------------

time_of_second = TimeCondWrapper(IsPeriod, TimeOfSecond)
time_of_minute = TimeCondWrapper(IsPeriod, TimeOfMinute)
time_of_hour = TimeCondWrapper(IsPeriod, TimeOfHour)
time_of_day = TimeCondWrapper(IsPeriod, TimeOfDay)
time_of_week = TimeCondWrapper(IsPeriod, TimeOfWeek)
time_of_month = TimeCondWrapper(IsPeriod, TimeOfMonth)

def every(past:str, based="run"):
    kws_past = {} # 'unit': 's'
    if based == "run":
        return TaskStarted(period=TimeDelta(past, kws_past=kws_past)) == 0
    elif based == "success":
        return TaskSucceeded(period=TimeDelta(past, kws_past=kws_past)) == 0
    elif based == "fail":
        return TaskFailed(period=TimeDelta(past, kws_past=kws_past)) == 0
    elif based == "finish":
        return TaskExecutable(period=TimeDelta(past, kws_past=kws_past))
    else:
        raise ValueError(f"Invalid status: {based}")

def cron(__expr=None, **kwargs):
    if __expr:
        args = __expr.split(" ")
    else:
        args = ()

    period = Cron(*args, **kwargs)
    return TaskRunnable(period=period)

def crontime(__expr=None, **kwargs):
    if __expr:
        args = __expr.split(" ")
    else:
        args = ()

    period = Cron(*args, **kwargs)
    return IsPeriod(period=period)

# Task pipelining
# ---------------

def after_success(task):
    return DependSuccess(depend_task=task)

def after_fail(task):
    return DependFailure(depend_task=task)

def after_finish(task):
    return DependFinish(depend_task=task)


def after_all_success(*tasks):
    return All(*(after_success(task) for task in tasks))

def after_all_fail(*tasks):
    return All(*(after_fail(task) for task in tasks))

def after_all_finish(*tasks):
    return All(*(after_finish(task) for task in tasks))


def after_any_success(*tasks):
    return Any(*(after_success(task) for task in tasks))

def after_any_fail(*tasks):
    return Any(*(after_fail(task) for task in tasks))

def after_any_finish(*tasks):
    return Any(*(after_finish(task) for task in tasks))

# Task Status
# -----------

def running(more_than:str=None, less_than=None, task=None):
    if more_than is not None or less_than is not None:
        period = TimeSpanDelta(near=more_than, far=less_than)
    else:
        period = None
    return TaskRunning(task=task, period=period)

retry = RetryWrapper()

started = TimeActionWrapper(TaskStarted)
failed = TimeActionWrapper(TaskFailed)
succeeded = TimeActionWrapper(TaskSucceeded)
finished = TimeActionWrapper(TaskFinished)

# Scheduler
# ---------

def scheduler_running(more_than:str=None, less_than:str=None):
    return SchedulerStarted(period=TimeSpanDelta(near=more_than, far=less_than))

def scheduler_cycles(more_than:int=None, less_than:int=None):
    kwds = {}
    if more_than is not None:
        kwds['__gt__'] = more_than
    if less_than is not None:
        kwds['__lt__'] = less_than
    return SchedulerCycles.from_magic(**kwds)



import logging
import pytest
from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo
from rocketry.core.log import TaskAdapter

def test_equal():
    logger = logging.getLogger("rocketry.task")
    try:
        logger.handlers = [RepoHandler(MemoryRepo())]
        assert TaskAdapter(logger, task="mytask") == TaskAdapter(logger, task="mytask")
        assert TaskAdapter(logger, task="mytask") != TaskAdapter(logger, task="another")
        with pytest.warns(UserWarning):
            assert TaskAdapter(logging.getLogger("rocketry._temp_test"), task="mytask") != TaskAdapter(logger, task="mytask")
    finally:
        logger.handlers = []

def test_add_handler():
    logger = logging.getLogger("rocketry._temp_test")
    try:
        hdlr = logging.StreamHandler()
        with pytest.warns(UserWarning):
            task_logger = TaskAdapter(logger, task="mytask")
        assert task_logger.handlers == []
        task_logger.addHandler(hdlr)

        assert task_logger.handlers == [hdlr]
        assert logger.handlers == [hdlr]
    finally:
        logger.handlers = []




import pytest

from rocketry.core import Parameters
from rocketry.args import Private

@pytest.mark.parametrize(
    "a,b,union",
    [
        pytest.param(
            lambda: Parameters({"a": 0, "b": 1}),
            lambda: Parameters({"c": 3, "d": 4}),
            lambda: Parameters({"a": 0, "b": 1, "c": 3, "d": 4}),
            id="Parameters | Parameters"),
        pytest.param(
            lambda: Parameters({"a": 0, "b": Private(1)}),
            lambda: Parameters({"c": 3, "d": Private(4)}),
            lambda: Parameters({"a": 0, "b": Private(1), "c": 3, "d": Private(4)}),
            id="Parameters | Parameters, with private"),
        pytest.param(
            lambda: Parameters({"a": 0, "b": 1}),
            lambda: {"c": 3, "d": 4},
            lambda: Parameters({"a": 0, "b": 1, "c": 3, "d": 4}),
            id="Parameters | Dict"),
        pytest.param(
            lambda: {"c": 3, "d": 4},
            lambda: Parameters({"a": 0, "b": 1}),
            lambda: Parameters({"a": 0, "b": 1, "c": 3, "d": 4}),
            id="Dict | Parameters", marks=pytest.mark.xfail),
    ],
)
def test_union(a, b, union):
    a = a()
    b = b()
    union = union()
    actual_union = a | b
    assert union == actual_union


@pytest.mark.parametrize(
    "a,b,union",
    [
        pytest.param(
            lambda: Parameters({"a": 0, "b": 1}),
            lambda: Parameters({"c": 3, "d": 4}),
            lambda: Parameters({"a": 0, "b": 1, "c": 3, "d": 4}),
            id="Parameters | Parameters"),
        pytest.param(
            lambda: Parameters({"a": 0, "b": Private(1)}),
            lambda: Parameters({"c": 3, "d": Private(4)}),
            lambda: Parameters({"a": 0, "b": Private(1), "c": 3, "d": Private(4)}),
            id="Parameters | Parameters, with private"),
        pytest.param(
            lambda: Parameters({"a": 0, "b": 1}),
            lambda: {"c": 3, "d": 4},
            lambda: Parameters({"a": 0, "b": 1, "c": 3, "d": 4}),
            id="Parameters | Dict"),
        pytest.param(
            lambda: {"c": 3, "d": 4},
            lambda: Parameters({"a": 0, "b": 1}),
            lambda: Parameters({"a": 0, "b": 1, "c": 3, "d": 4}),
            id="Dict | Parameters", marks=pytest.mark.xfail),
    ],
)
def test_update(a, b, union):
    a = a()
    b = b()
    union = union()
    a.update(b)
    assert union.materialize() == a.materialize()

def test_equal():
    assert Parameters({"a": 0, "b": 1}) == Parameters({"a": 0, "b": 1})
    assert not Parameters({"a": 0, "b": 1}) == 1
    assert Parameters({"a": 0, "b": 1}) != 1



import sys
import inspect
import importlib
from pathlib import Path
from typing import Callable, List, Optional
import warnings

from pydantic import Field, PrivateAttr, validator

from rocketry.core.task import Task
from rocketry.core.parameters import Parameters
from rocketry.pybox.pkg import find_package_root


def get_module(path, pkg_path=None):
    if pkg_path:
        name = '.'.join(
            path
            .with_suffix('') # path/to/file/myfile.py --> path.to.file.myfile
            .parts[len(pkg_path.parts):] # root/myproject/pkg/myfile.py --> myproject.pkg.myfile
        )
    else:
        name = Path(path).name

    spec = importlib.util.spec_from_file_location(name, path.absolute())
    task_module = importlib.util.module_from_spec(spec)

    try:
        spec.loader.exec_module(task_module)
    except Exception as exc:
        raise ImportError(f"Importing the file '{path}' failed.") from exc
    return task_module

class TempSysPath:
    # TODO: To utils.
    sys_path = sys.path
    def __init__(self, paths:list):
        self.paths = paths

    def __enter__(self):
        for path in self.paths:
            sys.path.append(path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for path in self.paths:
            try:
                self.sys_path.remove(path)
            except ValueError:
                pass

class FuncTask(Task):
    """Task that executes a function or callable.

    Parameters
    ----------
    func : Callable, str
        Function or name of a function to be executed. If string is
        passed, the path to the file where the function is should
        be passed with ``path`` or in the argument, like
        "path/to/file.py:my_func".
    path : path-like
        Path to the function. Not needed if ``func`` is callable.
    delay : bool, optional
        If True, the function is imported and set to the task
        immediately. If False, the function is imported only
        when running the task. By default False if ``func`` is
        callable and True if ``func`` is a name of a function.
    sys_path : list of paths
        Paths that are appended to ``sys.path`` when the function
        is imported.
    **kwargs : dict
        See :py:class:`rocketry.core.Task`


    Examples
    --------

    >>> from rocketry.tasks import FuncTask
    >>> def myfunc():
    ...     ...
    >>> task = FuncTask(myfunc, name="my_func_task_1")

    **Via decorator:**

    >>> from rocketry.tasks import FuncTask
    >>> @FuncTask(name='my_func_task_2', start_cond="daily")
    ... def myfunc():
    ...     ...

    If the ``name`` is not defined, the name will be in form
    ``path.to.module:myfunc``.

    Or from string using lazy importing:

    >>> from rocketry.tasks import FuncTask
    >>> task = FuncTask("myfunc", path="path/to/script.py", name='my_func_task_3', start_cond="daily")

    Warnings
    --------

    If ``execution='process'``, only picklable functions can be used.
    The following will NOT work:

    .. code-block:: python

        # Lambda functions are not allowed
        FuncTask(lambda:None, execution="process")

    .. code-block:: python

        # nested functions are not allowed
        def my_func():
            @FuncTask(execution="process")
            def my_task_func():
                ...
        my_func()

    .. code-block:: python

        def my_decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper

        # decorated functions are not allowed
        @my_decorator
        @FuncTask(execution="process")
        def my_task_func():
            ...
    """
    func: Optional[Callable] = Field(description="Executed function")

    path: Optional[Path] = Field(description="Path to the script that is executed")
    func_name: Optional[str] = Field(default="main", description="Name of the function in given path. Pass path as well")
    cache: bool = False

    sys_paths: List[Path] = []

    _is_delayed: bool = PrivateAttr(default=False)
    _delayed_kwargs: dict = {}
    _name_template: str = '{module_name}:{func_name}'
    @property
    def delayed(self):
        return self._is_delayed

    @validator('path')
    def validate_path(cls, value: Path, values):
        name = values['name']
        if value is not None and not value.is_file():
            warnings.warn(f"Path {value} does not exists. Task '{name}' may fail.")
        return value

    @validator("func")
    def validate_func(cls, value, values):
        execution = values.get('execution')
        func = value

        if execution == "process" and getattr(func, "__name__", None) == "<lambda>":
            raise AttributeError(
                f"Cannot pickle lambda function '{func}'. "
                "The function must be pickleable if task's execution is 'process'. "
            )
        return value

    def __init__(self, func=None, **kwargs):
        only_func_set = func is not None and not kwargs
        no_func_set = func is None and kwargs.get('path') is None
        if no_func_set:
            # FuncTask was probably called like:
            # @FuncTask(...)
            # def myfunc(...): ...

            # We initiate the class lazily by creating
            # almost empty shell class that is populated
            # in next __call__ (which should occur immediately)
            self._delayed_kwargs = kwargs
            return
        elif only_func_set:
            # Most likely called as:
            # @FuncTask
            # def myfunc(...): ...

            # We are slightly forgiving and set
            # the execution to else than process
            # as it's obvious it would not work.
            kwargs["execution"] = "thread"
        super().__init__(func=func, **kwargs)
        self._set_descr(is_delayed=func is None)

    def __call__(self, *args, **kwargs):
        if not hasattr(self, "func"):
            func = args[0]
            super().__init__(func=func, **self._delayed_kwargs)
            self._set_descr(is_delayed=False)
            self._delayed_kwargs = {}

            # Note that we must return the function or
            # we are in deep shit with multiprocessing
            # (or pickling the function).

            # As we return the function, the name of the
            # task might be missing. We set the name so
            # that condition API can identify the name.
            # If the task is renamed, the link is lost. (TODO)
            func.__rocketry__ = {'name': self.name}

            return func
        else:
            return super().__call__(*args, **kwargs)

    def _set_descr(self, is_delayed:bool):
        "Set description from func doc if desc missing"
        if self.description is None and hasattr(self.func, "__doc__"):
            self.description = self.func.__doc__
        # Set params
        if not is_delayed:
            # Not-delayed, setting parameters from the
            # function signature
            params = Parameters._from_signature(self.func)
            self.parameters = params | self.parameters
        self._is_delayed = is_delayed

    async def execute(self, **params):
        "Run the actual, given, task"
        func = self.get_func(cache=self.cache)

        is_async = inspect.iscoroutinefunction(func)
        if is_async:
            output = await func(**params)
        else:
            output = func(**params)
        return output

    def get_func(self, cache=True):
        if self.func is None:
            # Add dir of self.path to sys.path so importing from that dir works
            pkg_path = find_package_root(self.path)
            root = str(Path(self.path).parent.absolute()) if not pkg_path else str(pkg_path)

            # _task_func is cached to faster performance
            with TempSysPath([root] + self.sys_paths):
                task_module = get_module(self.path, pkg_path=pkg_path)
            task_func = getattr(task_module, self.func_name)

            if cache:
                self.func = task_func
            return task_func
        else:
            return self.func

    def get_default_name(self, func=None, path=None, func_name=None, _name_template=None, **kwargs):
        if func is None:
            file = Path(path)
            module_name = '.'.join(file.parts).replace(".py", "")
        else:
            module_name = func.__module__
            func_name = getattr(func, "__name__", type(func).__name__)
            if module_name == "__main__":
                # Showing as 'myfunc'
                return func_name
        if _name_template is not None:
            return _name_template.format(module_name=module_name, func_name=func_name)
        else:
            return f'{module_name}:{func_name}'

    def process_finish(self, *args, **kwargs):
        if self._is_delayed:
            # Deleting the _func so it is refreshed
            # next time the task is run.
            self.func = None
        super().process_finish(*args, **kwargs)

    def is_delayed(self):
        return self._is_delayed

    def get_task_params(self):
        task_params = super().get_task_params()

        if self._is_delayed:
            # Get params from the typehints
            cache = False if self.path is not None else True
            func = self.get_func(cache=cache)
            func_params = Parameters._from_signature(func, task=self, session=self.session)
            params = func_params | task_params
        else:
            params = task_params
        return params

    def prefilter_params(self, params):
        if not self.is_delayed():
            # Filter the parameters now so that
            # we pass as little as possible to
            # pickling. If lazy, we filter after
            # pickling to handle problems in
            # pickling functions.
            return {
                key: val for key, val in params.items()
                if key in self.kw_args
            }
        else:
            return params

    def postfilter_params(self, params:Parameters):
        if self._is_delayed:
            # Was not filtered in prefiltering.
            return {
                key: val for key, val in params.items()
                if key in self.kw_args
            }
        else:
            return params

    @property
    def pos_args(self):
        func = self.get_func()
        sig = inspect.signature(func)
        pos_args = [
            val.name
            for name, val in sig.parameters.items()
            if val.kind in (
                inspect.Parameter.POSITIONAL_ONLY, # NOTE: Python <= 3.8 do not have positional arguments, but maybe in the future?
                inspect.Parameter.POSITIONAL_OR_KEYWORD # Keyword argument
            )
        ]
        return pos_args

    @property
    def kw_args(self):
        func = self.get_func()
        sig = inspect.signature(func)
        kw_args = [
            val.name
            for name, val in sig.parameters.items()
            if val.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD, # Normal argument
                inspect.Parameter.KEYWORD_ONLY # Keyword argument
            )
        ]
        return kw_args




import pytest

from rocketry.conditions import (
    All, Any, Not
)
from rocketry.conds import true, false

@pytest.mark.parametrize("actual,expected",
    [
        pytest.param(
            lambda: (true & (true & true)),
            lambda: All(true, true, true),
            id="All"
        ),
        pytest.param(
            lambda: (true | (true | true)),
            lambda: Any(true, true, true),
            id="Any"
        ),
        pytest.param(
            lambda: Not(Not(Not(true))),
            lambda: Not(Not(Not(true))),
            id="Not"
        ),
        pytest.param(
            lambda: (true & true) & (false | false),
            lambda: All(All(true, true), Any(false, false)),
            id="Complex"
        ),
    ]
)
def test_nested(actual, expected):
    expected = expected()
    actual = actual()
    assert expected == actual



import logging
import pytest

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo

from rocketry.conditions.scheduler import SchedulerStarted
from rocketry.core.time.base import TimeDelta
from rocketry.log.log_record import LogRecord
from rocketry.tasks import FuncTask
from rocketry.conditions import TaskStarted, AlwaysTrue

def run_failing():
    raise RuntimeError("Task failed")

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_fail_traceback(tmpdir, execution, session):
    # There is a speciality in tracebacks in multiprocessing
    # See: https://bugs.python.org/issue34334

    # TODO: Delete. This has been handled now in test_core.py
    with tmpdir.as_cwd() as old_dir:
        task_logger = logging.getLogger(session.config.task_logger_basename)
        task_logger.handlers = [
            RepoHandler(repo=MemoryRepo(model=LogRecord))
        ]
        task = FuncTask(run_failing, name="task", start_cond=AlwaysTrue(), execution=execution, session=session)

        session.config.shut_cond = (TaskStarted(task="task") >= 3) | ~SchedulerStarted(period=TimeDelta("20 seconds"))
        session.start()

        failures = list(task.logger.get_records(action="fail"))
        assert 3 == len(failures)

        for record in failures:
            tb = record.exc_text
            assert "Traceback (most recent call last):" in tb
            assert "RuntimeError: Task failed" in tb



import datetime
import logging

import pytest

from rocketry.conditions import (
    TaskExecutable,
)
from rocketry.pybox.time.convert import to_datetime
from rocketry.time import (
    TimeOfDay
)
from rocketry.tasks import FuncTask
from rocketry.time.interval import TimeOfMinute
from rocketry.testing.log import create_task_record

@pytest.mark.parametrize("from_logs", [pytest.param(True, id="from logs"), pytest.param(False, id="optimized")])
@pytest.mark.parametrize(
    "get_condition,logs,time_after,outcome",
    [
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-01 07:30",
            False,
            id="Don't run (already succeeded)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "fail"),
            ],
            "2020-01-01 07:30",
            False,
            id="Don't run (already failed)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "terminate"),
            ],
            "2020-01-01 07:30",
            False,
            id="Don't run (terminated)"),

        pytest.param(
            # Termination is kind of failing but retry is not applicable as termination is often
            # indication that the task is not desired to be run anymore (as it has already taken
            # enough system resources)
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00"), retries=1),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "terminate"),
            ],
            "2020-01-01 07:30",
            False,
            id="Don't run (terminated with retries)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-01 08:30",
            False,
            id="Don't run (already ran and out of time)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2021-12-31 08:30",
            False,
            id="Don't run (missed)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-02 06:00",
            False,
            id="Don't run (not yet time)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [],
            "2020-01-01 08:30",
            False,
            id="Don't run (out of time and not run at all)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "inaction"),
            ],
            "2020-01-01 07:30",
            False,
            id="Do not run (has inacted)"), # One would not want the task to try to run every millisecond if inacted.

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
            ],
            "2020-01-01 07:30",
            False,
            id="Do not run (is already running)", marks=pytest.mark.xfail(reason="Bug but does not affect if no multiple schedulers nor multilaunch allowed")), # One would not want the task to try to run every millisecond if inacted.

        # Do run
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [],
            "2020-01-01 07:10",
            True,
            id="Do run (has not run at all)"),
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00"), retries=1),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "fail"),
            ],
            "2020-01-01 07:30",
            True,
            id="Do run (has retries)"),
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-02 07:30",
            True,
            id="Do run (succeeded yesterday)"),
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "fail"),
            ],
            "2020-01-02 07:30",
            True,
            id="Do run (failed yesterday)"),
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "terminate"),
            ],
            "2020-01-02 07:30",
            True,
            id="Do run (terminated yesterday)"),

        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "inaction"),
            ],
            "2020-01-02 07:30",
            True,
            id="Do run (inacted yesterday)"),

    ],
)
def test_executable(tmpdir, mock_datetime_now, logs, time_after, get_condition, outcome, session, from_logs):
    session.config.force_status_from_logs = from_logs
    def to_epoch(dt):
        # Hack as time.tzlocal() does not work for 1970-01-01
        if dt.tz:
            dt = dt.tz_convert("utc").tz_localize(None)
        return (dt - datetime.datetime(1970,  1, 1)) // datetime.timedelta(seconds=1)

    with tmpdir.as_cwd() as old_dir:


        task = FuncTask(
            lambda:None,
            name="the task",
            execution="main",
            session=session
        )

        condition = get_condition()

        for log in logs:
            log_time, log_action = log[0], log[1]
            log_created = to_datetime(log_time).timestamp()
            record = create_task_record(
                msg="Logging of 'task'", args=(),
                task_name="the task", created=log_created, action=log_action
            )

            task.logger.handle(record)
            setattr(task, f'last_{log_action}', to_datetime(log_time))
        mock_datetime_now(time_after)

        if outcome:
            assert condition.observe(session=session)
            assert condition.observe(task=task)
        else:
            assert not condition.observe(session=session)
            assert not condition.observe(task=task)


@pytest.mark.parametrize("from_logs", [pytest.param(True, id="from logs"), pytest.param(False, id="optimized")])
@pytest.mark.parametrize(
    "get_condition,logs,time_after,outcome",
    [
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay()),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-02 00:01",
            True,
            id="Do run (continuous cycle, daily)"),
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay() & TimeOfMinute()),
            [
                ("2020-01-01 23:59", "run"),
                ("2020-01-01 23:59", "success"),
            ],
            "2020-01-02 00:01",
            True,
            id="Do run (continuous cycle, daily & minutely)"),
        pytest.param(
            lambda:TaskExecutable(task="the task", period=TimeOfDay() | TimeOfMinute()),
            [
                ("2020-01-01 23:59", "run"),
                ("2020-01-01 23:59", "success"),
            ],
            "2020-01-02 00:01",
            True,
            id="Do run (continuous cycle, daily | minutely)"),
    ]
)
def test_periods(mock_datetime_now, logs, time_after, get_condition, outcome, session, from_logs):
    "Sanity check that periods work correctly"
    session.config.force_status_from_logs = from_logs

    task = FuncTask(
        lambda:None,
        name="the task",
        execution="main",
        session=session
    )

    condition = get_condition()

    for log in logs:
        log_time, log_action = log[0], log[1]
        log_created = to_datetime(log_time).timestamp()
        record = logging.LogRecord(
            # The content here should not matter for task status
            name='rocketry.core.task', level=logging.INFO, lineno=1,
            pathname='d:\\Projects\\rocketry\\rocketry\\core\\task\\base.py',
            msg="Logging of 'task'", args=(), exc_info=None,
        )

        record.created = log_created
        record.action = log_action
        record.task_name = "the task"

        task.logger.handle(record)
        setattr(task, f'last_{log_action}', to_datetime(log_time))
    mock_datetime_now(time_after)

    if outcome:
        assert condition.observe(session=session)
        assert condition.observe(task=task)
    else:
        assert not condition.observe(session=session)
        assert not condition.observe(task=task)



import datetime

import pytest

from rocketry.conditions import (
    TaskRunnable,
)
from rocketry.pybox.time.convert import to_datetime
from rocketry.time import (
    TimeOfDay
)
from rocketry.tasks import FuncTask
from rocketry.testing.log import create_task_record

@pytest.mark.parametrize("from_logs", [pytest.param(True, id="from logs"), pytest.param(False, id="optimized")])
@pytest.mark.parametrize(
    "get_condition,logs,time_after,outcome",
    [
        pytest.param(
            lambda:TaskRunnable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-01 07:30",
            False,
            id="Don't run (already ran)"),

        pytest.param(
            lambda:TaskRunnable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-01 08:30",
            False,
            id="Don't run (already ran, out of the period)"),

        pytest.param(
            lambda:TaskRunnable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [],
            "2020-01-01 08:30",
            False,
            id="Don't run (out of the period)"),

        pytest.param(
            lambda:TaskRunnable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [],
            "2020-01-02 07:30",
            True,
            id="Do run (has not run)"),

        pytest.param(
            lambda:TaskRunnable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 07:10", "run"),
                ("2020-01-01 07:20", "inaction"),
            ],
            "2020-01-02 07:30",
            True,
            id="Do run (ran yesterday)"),

        pytest.param(
            lambda:TaskRunnable(task="the task", period=TimeOfDay("07:00", "08:00")),
            [
                ("2020-01-01 06:50", "run"),
                ("2020-01-01 07:20", "success"),
            ],
            "2020-01-01 07:30",
            True,
            id="Do run (ran outside the period)"),

    ],
)
def test_runnable(tmpdir, mock_datetime_now, logs, time_after, get_condition, outcome, session, from_logs):
    session.config.force_status_from_logs = from_logs
    def to_epoch(dt):
        # Hack as time.tzlocal() does not work for 1970-01-01
        if dt.tz:
            dt = dt.tz_convert("utc").tz_localize(None)
        return (dt - datetime.datetime(1970,  1, 1)) // datetime.timedelta(seconds=1)

    with tmpdir.as_cwd() as old_dir:


        task = FuncTask(
            lambda:None,
            name="the task",
            execution="main",
            session=session
        )

        condition = get_condition()

        for log in logs:
            log_time, log_action = log[0], log[1]
            record = create_task_record(
                # The content here should not matter for task status
                name='rocketry.core.task', lineno=1,
                pathname='d:\\Projects\\rocketry\\rocketry\\core\\task\\base.py',
                msg="Logging of 'task'", args=(), exc_info=None,
                created=log_time, action=log_action, task_name="the task"
            )

            task.logger.handle(record)
            setattr(task, f'last_{log_action}', to_datetime(log_time))
        mock_datetime_now(time_after)

        if outcome:
            assert condition.observe(session=session)
            assert condition.observe(task=task)
        else:
            assert not condition.observe(session=session)
            assert not condition.observe(task=task)




from datetime import datetime, timedelta
import pytest
from rocketry.core.time import (
    TimeDelta
)

@pytest.mark.parametrize(
    "dt,dt_ref,offset",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            datetime(2020, 1, 1, 12, 00),
            "2:00:00",
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00),
            datetime(2020, 1, 1, 12, 00),
            "2:00:00",
            id="Right of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 00),
            datetime(2020, 1, 1, 12, 00),
            "2:00:00",
            id="Middle of interval"),
    ],
)
def test_in_offset(dt, dt_ref, offset):
    time = TimeDelta(offset, reference=dt_ref)
    assert dt in time

@pytest.mark.parametrize(
    "dt,dt_ref,offset",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 9, 59, 59, 999999),
            datetime(2020, 1, 1, 12, 00),
            "2:00:00",
            id="Left from interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00, 1),
            datetime(2020, 1, 1, 12, 00),
            "2:00:00",
            id="Right from interval"),
    ],
)
def test_not_in_offset(dt, dt_ref, offset):
    time = TimeDelta(offset, reference=dt_ref)
    assert dt not in time


def test_reference_now():
    time = TimeDelta("10 seconds")
    assert datetime.now() in time
    assert (datetime.now() - timedelta(0, 5, 0)) in time
    assert (datetime.now() - timedelta(0, 11, 0)) not in time



from rocketry.conditions.task import DependFailure, DependSuccess
from rocketry.core.condition.base import All, Any
from rocketry.tasks import FuncTask
from rocketry.utils.dependencies import Dependencies, Link, get_dependencies

def test_dependency(session):
    ta = FuncTask(lambda: None, name="a", start_cond="daily", execution="main", session=session)
    tb = FuncTask(lambda: None, name="b", start_cond="after task 'a'", execution="main", session=session)
    tc = FuncTask(lambda: None, name="c", start_cond="after task 'a' & after task 'b' failed", execution="main", session=session)
    td = FuncTask(lambda: None, name="d", start_cond="after task 'a' | after task 'b'", execution="main", session=session)

    te = FuncTask(lambda: None, name="no link", start_cond="daily", execution="main", session=session)

    # Test Dependencies
    # -----------------
    deps = Dependencies(session)
    deps_list = sorted(deps, key=lambda x: (x.child.name, x.parent.name))
    assert list(deps_list) == [
        Link(ta, tb, relation=DependSuccess),
        Link(ta, tc, relation=DependSuccess, type=All),
        Link(tb, tc, relation=DependFailure, type=All),
        Link(ta, td, relation=DependSuccess, type=Any),
        Link(tb, td, relation=DependSuccess, type=Any),
    ]
    assert str(deps)
    assert repr(deps)

    # Test get_dependencies
    # ---------------------
    deps = get_dependencies(session)
    deps_list = sorted(deps, key=lambda x: (x.child.name, x.parent.name))
    assert deps_list == [
        Link(ta, tb, relation=DependSuccess),
        Link(ta, tc, relation=DependSuccess, type=All),
        Link(tb, tc, relation=DependFailure, type=All),
        Link(ta, td, relation=DependSuccess, type=Any),
        Link(tb, td, relation=DependSuccess, type=Any),
    ]

def test_link(session):
    ta = FuncTask(lambda: None, name="a", start_cond="daily", execution="main", session=session)
    tb = FuncTask(lambda: None, name="b", start_cond="after task 'a'", execution="main", session=session)
    tc = FuncTask(lambda: None, name="c", start_cond="after task 'a' & after task 'b' failed", execution="main", session=session)
    td = FuncTask(lambda: None, name="d", start_cond="after task 'a' | after task 'b'", execution="main", session=session)

    assert str(Link(ta, tb, relation=DependSuccess)) == "'a' -> 'b'"
    assert str(Link(ta, tb, relation=DependFailure)) == "'a' -> 'b'"
    assert str(Link(ta, tb, relation=DependSuccess, type=Any)) == "'a' -> 'b'"
    assert str(Link(ta, tb, relation=DependSuccess, type=All)) == "'a' -> 'b' (multi)"

    assert repr(Link(ta, tb, relation=DependSuccess, type=All)) == "Link('a', 'b', relation=DependSuccess, type=All)"




import os
import sys
import warnings
from typing import Any, Callable, Optional
try:
    from typing import Literal
except ImportError: # pragma: no cover
    from typing_extensions import Literal

from rocketry.core.parameters import BaseArgument, Parameters

class NotSet:
    def __repr__(self):
        return 'NOTSET'

NOTSET = NotSet()

class SimpleArg(BaseArgument):
    """A simple argument.

    Parameters
    ----------
    value : Any
        Value of the argument.

    """
    def __init__(self, value:Any):
        self._value = value

    def get_value(self, task=None, **kwargs) -> Any:
        return self._value

class Arg(BaseArgument):
    """A simple argument got from the session

    Parameters
    ----------
    value : Any
        Value of the argument.

    """
    def __init__(self, key:Any, default=NOTSET):
        self.key = key
        self.default = default

    def get_value(self, task=None, session=None, **kwargs) -> Any:
        if session is None:
            session = task.session
        try:
            return session.parameters._get(self.key, task=task, session=session, **kwargs)
        except KeyError:
            if self.default is NOTSET:
                raise
            return self.default

    def __str__(self):
        return self.key

    def __repr__(self):
        return f'session.parameters[{repr(self.key)}]'

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.key == other.key
        else:
            return False

class Session(BaseArgument):
    "An argument that represents the session"

    def get_value(self, task=None, session=None, **kwargs) -> Any:
        if session is not None:
            return session
        else:
            return task.session

    def __repr__(self):
        return f'session'

class Task(BaseArgument):
    "An argument that represents a task"

    def __init__(self, name=None):
        self.name = name

    def get_value(self, task=None, **kwargs) -> Any:
        if self.name is None:
            return task
        else:
            return task.session[self.name]

    def __repr__(self):
        return f'Task({repr(self.name) if self.name is not None else ""})'

class Return(BaseArgument):
    """A return argument

    Return is an argument which value represents
    the return of a task.

    Parameters
    ----------
    name : str
        Name of the task which return value the argument
        represents.

    Examples
    --------

    .. code-block:: python

        from rocketry.arguments import Return
        from rocketry.tasks import FuncTask

        # Create a task with a return
        @FuncTask()
        def my_task_1():
            ...
            return data

        @FuncTask(parameters={"myarg": Return('my_task_1')})
        def my_task_2(myarg):
            ...
    """

    def __init__(self, task_name, default=NOTSET):
        self.task_name = task_name
        self.default = default

    def get_value(self, task=None, session=None, **kwargs) -> Any:
        if session is None:
            session = task.session
        try:
            input_task = session[self.task_name]
        except KeyError:
            raise ValueError(f"Task {repr(self.task_name)} does not exists. Cannot get return value")
        try:
            return session.returns[input_task]
        except KeyError:
            if self.default is NOTSET:
                raise KeyError(f"Return value not found for {repr(task)}")
            return self.default

    def __repr__(self):
        return f'Return({repr(self.task_name)}{"" if self.default is None else ", default=" + repr(self.default)})'

    def __str__(self):
        return f'Return of {self.task_name!r}'

class FuncArg(BaseArgument):
    """An argument which value is defined by the
    return value of given function.

    Parameters
    ----------
    func : Callable
        Function that is executed and passed as value.
    *args : tuple
        Positional arguments passed to func.
    **kwargs : dict
        Keyword arguments passed to func.

    Examples
    --------

    Simple example:

    .. code-block:: python

        from rocketry.tasks import FuncTask
        from rocketry.arguments import FuncArg

        def my_func():
            ...
            return obj

        @FuncTask(parameters={"my_param": FuncArg(my_func)})
        def my_task(my_param):
            ...

    Example to set to session:

    .. doctest:: funcarg

        >>> from rocketry.arguments import FuncArg

    The ``session.parameters`` were updated.

    .. doctest:: funcarg

        >>> from rocketry import session
        >>> session.parameters
        Parameters(myarg1=FuncArg(myarg1), myarg2=FuncArg(myfunc))
    """
    def __init__(self, __func:Callable, *args, materialize:Optional[Literal['pre', 'post']]=None, **kwargs):
        self.func = __func
        self.materialize = materialize
        self.args = args
        self.kwargs = kwargs

    def get_value(self, **kwargs):
        return self(**kwargs)

    def __call__(self, **kwargs):
        param_kwargs = Parameters._from_signature(self.func)
        params = param_kwargs.materialize(**kwargs)
        return self.func(*self.args, **params, **self.kwargs)

    def stage(self, **kwargs):
        session = kwargs['session']
        materialize = self.materialize if self.materialize is not None else session.config.param_materialize

        if materialize == "pre":
            return self.get_value(**kwargs)
        else:
            return self

    def __repr__(self):
        cls_name = type(self).__name__
        return f'{cls_name}({self.func.__name__})'

class TerminationFlag(BaseArgument):

    def get_value(self, task=None, session=None, **kwargs) -> Any:
        execution = task.execution
        if execution in ("process", "main"):
            warnings.warn(f"Passing termination flag to task with 'execution_type={execution}''. Flag cannot be used.")
        return task._thread_terminate

    def __repr__(self):
        return 'TerminationFlag()'

    def __str__(self):
        return 'termination flag'

class EnvArg(BaseArgument):
    """Argument that has the value of an environment variable"""

    def __init__(self, var, default=NOTSET):
        self.var = var
        self.default = default

    def get_value(self, **kwargs) -> Any:
        try:
            return os.environ[self.var]
        except KeyError:
            if self.default is NOTSET:
                raise
            return self.default

class CliArg(BaseArgument):
    """Argument that has the value of a command line argument"""

    cli_args = sys.argv

    def __init__(self, param, default=NOTSET):
        self.param = param
        self.default = default

    def get_value(self, **kwargs) -> Any:
        return self._get_arg(self.cli_args)

    def _get_arg(self, args:list):
        arg_name = self.param
        for i, arg in enumerate(args):
            if arg == arg_name:
                break
        else:
            if self.default is NOTSET:
                raise KeyError("CLI argument not found")
            else:
                return self.default
        return args[i+1]



import warnings

import pytest
from rocketry import Session

def test_shutdown(session):
    assert not session.scheduler._flag_shutdown.is_set()
    with pytest.warns(DeprecationWarning):
        session.shutdown()
    assert session.scheduler._flag_shutdown.is_set()

def test_no_execution_method():

    with pytest.warns(FutureWarning):
        Session()

    with warnings.catch_warnings():
        warnings.simplefilter("error")

        # Test the following won't warn
        Session(config=dict(task_execution="process"))
        Session(config=dict(task_execution="thread"))
        Session(config=dict(task_execution="main"))
        Session(config=dict(task_execution="async"))



import pytest
from rocketry.args import SimpleArg, Arg, EnvArg, CliArg

def test_equal():
    assert SimpleArg("A") != SimpleArg("B")
    assert not SimpleArg("A") == SimpleArg("B")
    assert SimpleArg("A") == SimpleArg("A")

def test_pipeline(session):
    session.parameters['found'] = "a value"
    session.parameters['not_reached'] = "a value 2"

    comb_arg = CliArg("--this-is-missing") >> EnvArg('ROCKETRY_MISSING') >> Arg('missing') >> Arg('found') >> Arg('not_reached')
    assert comb_arg.get_value(session=session) == "a value"

    comb_arg = CliArg("--this-is-missing", default="a value 2") >> EnvArg('ROCKETRY_MISSING') >> Arg('missing') >> Arg('found') >> Arg('not_reached')
    assert comb_arg.get_value(session=session) == "a value 2"

    with pytest.raises(KeyError):
        (CliArg("--this-is-missing") >> EnvArg('ROCKETRY_MISSING')).get_value(session=session)
    with pytest.raises(TypeError):
        CliArg("--this-is-missing", default="a value 2") >> 4



from logging.handlers import QueueHandler as _QueueHandler
from logging import Formatter

import copy

# Copying the default formatter mechanism from logging
_DEFAULT_FORMATTER = Formatter()

class QueueHandler(_QueueHandler):
    """
    The logging.handlers.QueueHandler gives priority for the message but does
    not maintain the other info in a record when passing it to the queue.
    These
    """

    def prepare(self, record):
        """
        Prepare a record for queuing. The object returned by this method is
        enqueued.

        The base implementation formats the record to merge the message and
        arguments, and removes unpickleable items from the record in-place.
        Specifically, it overwrites the record's `msg` and
        `message` attributes with the merged message (obtained by
        calling the handler's `format` method), and sets the `args`,
        `exc_info` and `exc_text` attributes to None.

        You might want to override this method if you want to convert
        the record to a dict or JSON string, or send a modified copy
        of the record while leaving the original intact.
        """
        # The format operation gets traceback text into record.exc_text
        # (if there's exception data), and also returns the formatted
        # message. We can then use this to replace the original
        # msg + args, as these might be unpickleable. We also zap the
        # exc_info and exc_text attributes, as they are no longer
        # needed and, if not None, will typically not be pickleable.
        msg = self.format(record)
        # bpo-35726: make copy of record to avoid affecting other handlers in the chain.
        record = copy.copy(record)
        record.message = msg
        # record.msg = msg
        record.args = None
        record.exc_info = None
        # record.exc_text = None
        return record
